from __future__ import print_function, division
from PyAstronomy.pyaC import ImportCheck as _ImportCheck
from PyAstronomy.pyaC import pyaErrors as _PE
import PyAstronomy as _PyA
import os, sys
from six.moves import configparser as _ConfigParser
import numpy
import six

_ic = _ImportCheck(["quantities"], ["quantities"])

import quantities as _pq

_systems = ["SI", "cgs"]
_unitSystem = "cgs"
inventory = {}

from .itc import PyAConstants, _displaySummary



def summary():
  """
    Print a summary of available constants to screen.
  """
  _displaySummary(sys.modules[__name__])

    
def constantDetails(const):
  """
    Print all information for a specific constant.
    
    Parameters
    ----------
    const : string
        The constant.
    
    Returns
    -------
    info : dictionary
        A dictionary containing all information about the
        specified constant.
  """
  if not const in inventory:
    raise(_PE.PyAValError("No such constant: '" + str(const) + "'."))
  cmod = sys.modules[__name__]
  print("Constant              : '" + const + "'")
  print("Description           : " + inventory[const]["descr"])
  print("Current value and unit: " + str(getattr(cmod, "f_"+const)))
  print("Uncertainty           : " + str(getattr(cmod, "f_"+const+"_err")))
  print("Source                : " + inventory[const]["source"])
  return inventory[const]

def inUnitsOf(const, units):
  """
    Get a constant and apply unit conversion.
    
    Returns
    -------
    Constant : Quantity object (`quantities` package)
        The value and unit of the constant.
  """
  cmod = sys.modules[__name__]
  if not hasattr(cmod, "f_" + const):
    raise(_PE.PyAValError("No such constant: '" + str(const) + "'"))
  val = getattr(cmod, "f_" + const)
  return val.rescale(units)

def _inventoryToModule():
  """
    Translate information in `inventory` into module attributes.
  """
  cmod = sys.modules[__name__]
  for k, u in six.iteritems(inventory):
    val = _pq.Quantity(numpy.float64(u["valueSI"]), u["units"]["SI"])
    value = val.rescale(u["units"][_unitSystem])
    setattr(cmod, u["symbol"], value.magnitude)
    setattr(cmod, "f_" + u["symbol"], value)
    err = _pq.Quantity(numpy.float64(u["errSI"]), u["units"]["SI"])
    err = err.rescale(u["units"][_unitSystem])
    setattr(cmod, "f_" + u["symbol"] + "_err", err)

def cleanUp():
  """
    Clean up loaded constants.
    
    Empties the `inventory` and deletes all
    related module attributes.
  """
  global inventory
  cmod = sys.modules[__name__]
  for k, u in six.iteritems(inventory):
    delattr(cmod, u["symbol"])
    delattr(cmod, "f_" + u["symbol"])
    delattr(cmod, "f_" + u["symbol"] + "_err")
  inventory = {}

def setSystem(system):
  """
    Set the unit system.
    
    Parameters
    ----------
    system : string, {SI, cgs}
        The unit system in use.
  """
  global _unitSystem
  if not system in _systems:
    raise(_PE.PyAValError("No such system: '" + system + "'.", \
                          solution="Choose one of: "+str(_systems)))
  _unitSystem = system
  _inventoryToModule()

def getSystem():
  """
    Get current unit system.
    
    Returns
    -------
    System : string, {SI, erg}
        The unit system currently in use.
  """
  global _unitSystem
  return _unitSystem

def load(fn, constClobber=False):
  """
    Load information about constants from file.
    
    A file containing information about constants is a
    simple Python configuration file. Each constant
    must be described in its own (arbitrarily named)
    section.
    
    An example for a valid entry looks like this: 
    
    ::
    
      [NewtonGravitation]
      descr   = Newton's constant of gravitation
      symbol  = G
      valueSI = 6.67428e-11 
      errSI   = 6.7e-15
      unitSI  = m^3 * kg^-1 * s^-2
      unitcgs = cm^3 * g^-1 * s^-2
      source  = http://maia.usno.navy.mil/NSFA/IAU2009_consts.html
    
    Note that all fields are mandatory.
    
    The units (and conversion) are managed by the `quantities`
    package.
    
    Parameters
    ----------
    fn : string
        Name of the file.
    constClobber : boolean, optional
        If this is set to true, previously loaded constants will be
        overwritten by newly loaded definitions. Otherwise, an error
        will be issued.
  """
  global inventory
  config = _ConfigParser.RawConfigParser()
  config.read(fn)
  for s in config.sections():
    nc = None
    try:
      nc = {}
      nc["symbol"]  = config.get(s, "symbol")
      nc["descr"]   = config.get(s, "descr")
      nc["valueSI"] = config.get(s, "valueSI")
      nc["errSI"]   = config.get(s, "errSI")
      nc["units"] = {}
      for u in _systems:
        nc["units"][u] = config.get(s, "unit"+u)
      nc["source"] = config.get(s, "source")
    except _ConfigParser.NoOptionError as e:
      _PE.warn(_PE.PyAValError("The definition of the constant defined in section '" + s + "' of file '" + \
                               fn + "' is incomplete. Ignoring this constant. Caught the error:\n" + \
                               str(e)))
      nc = None
    if nc is not None:
      if (nc["symbol"] in inventory) and (not constClobber):
        raise(_PE.PyANameClash("A constant with the symbol '" + nc["symbol"] + "' is already defined.", \
                               solution=["Choose another symbol.", "Set the `constClobber` flag to True to overwrite existing definitions of constants."])) 
      inventory[nc["symbol"]] = nc
    _inventoryToModule()



load(os.path.join(os.path.dirname(_PyA.__file__), "constants/cdat.dat"))



import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Model object and parameter access
        # =============================
        
        # In[1]:
        
        
        from PyAstronomy import funcFit2 as fuf2
        import matplotlib
        matplotlib.use('Agg')
        
        # Create a a model object representing a Gaussian
        gf = fuf2.GaussFit()
        
        # Get some information on the available parameters
        gf.parameterSummary()
        
        # Parameters can be accessed using the square bracket
        # notation
        gf["A"] = -10.0
        gf["sig"] = 15.77
        gf["off"] = 0.96
        gf["mu"] = 7.5
        print()
        print("Parameters values: ")
        print("  A   : ", gf["A"])
        print("  sig : ", gf["sig"])
        print("  off : ", gf["off"])
        print("  mu  : ", gf["mu"])
        print()
        
        # Wildcards can be used to specify parameters
        gf["*"] = 0.75
        gf["o?f"] = 1.0 
        
        # More convenient overview of parameter status
        gf.parameterSummary()
        print()
        
        # Exporting and assigning parameter values via dictionary
        ps = gf.parameters()
        print("Parameter name/value dictionary: ", ps)
        # Assigning values from dictionary (identical values here)
        gf.assignValues(ps)
        



from pathlib import Path
from nbconvert import ScriptExporter
import nbformat


def insert_mpl_agg(s):
    # Find last import ...
    x = 0
    for i, l in enumerate(s):
        if l.find("import") != -1:
            x = i
    # ... and import matplotlib to set backend to Agg
    s.insert(x+1,"matplotlib.use('Agg')")
    s.insert(x+1,"import matplotlib")
    return s

def del_future(s):
    # Remove __future__ import
    for i, l in enumerate(s):
        if l.find("from __future__") != -1:
            s.pop(i)
            return s
    return s


def sanitize(s, name="example"):
    x = """
import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_%s(self):""" % name
    x = x.split("\n")
    s = [" "*8+l for l in s]
    x.extend(s)
    return x


fns = list(Path(".").rglob("ex_*.ipynb"))

print("Files matching criteria:")
for i, fn in enumerate(fns,1):
    print(i, fn)
print()

for fn in fns:
    if str(fn).find("_build") != -1:
        continue
    c = ''.join(open(str(fn), 'rt').readlines())
    nb = nbformat.reads(c, as_version=4)
    se = ScriptExporter()
    (body, resources) = se.from_notebook_node(nb)
    s = body.split("\n")
    
    s = del_future(s)
    s = insert_mpl_agg(s)
    s = sanitize(s)
    
    ofp = fn.with_name("sanity_"+fn.stem+".py")
    print("Input notebook: ", fn)
    print("Output test: ", ofp)
    
    open(str(ofp), 'wt').writelines([l+"\n" for l in s])
    



import numpy as np

def idlMod(a, b):
  """
    Emulate 'modulo' behavior of IDL.
    
    Parameters
    ----------
    a : float or array
        Numerator
    b : float
        Denominator
    
    Returns
    -------
    IDL modulo : float or array
        The result of IDL modulo operation.
  """
  if isinstance(a, np.ndarray):
    s = np.sign(a)
    m = np.mod(a, b)
    m[(s < 0)] -= b
  else:
    m = a % b
    if a < 0: m -= b
  return m


import numpy as np
from PyAstronomy import pyaC as PC
from PyAstronomy.pyaC import pyaErrors as PE


def airmassPP(zangle):
    """
    Calculate airmass for plane parallel atmosphere.

    Parameters
    ----------
    zangle : float or array
        The zenith angle in degrees.

    Returns
    -------
    Airmass : float or array
        The airmass assuming a plane parallel
        atmosphere.
    """
    return 1.0 / np.cos(PC.degtorad(zangle))


def airmassSpherical(zangle, obsAltitude, rearth=6371.0, yatm=10.0):
    """
    Calculate the airmass for a given zenith angle and observer altitude.

    This routine uses a geometric formula for a homogeneous, spherical
    atmosphere with an elevated observer.

    .. note:: In this model, the airmass is *not* necessarily one
              toward the zenith.

    Parameters
    ----------
    zangle : float, array
        Zenith angle of an object in deg.
    obsAltitude :  float
        Elevation of the observer in meter.
    rearth : float, optional
        Earth's radius in km.
    yatm : float, optional
        Height of the atmosphere in km.

    Returns
    -------
    Airmass : float or array
        The airmass.
    """

    # Convert observer's altitude to km to have
    # consistent units
    obsAltitude = obsAltitude / 1000.0
    r = (rearth / yatm)
    y = (obsAltitude / yatm)

    # Find maximal zenith angle
    zmax = 180.0 - PC.radtodeg(np.arcsin(rearth / (rearth + obsAltitude)))
    if np.any(zangle > zmax):
        raise(PE.PyAValError("At least one zenith angle is too large. The maximum allowed angle is " + str(zmax) + " deg.",
                             solution="Use an angle within the limit. Check observer's altitude."))

    # Convert into deg
    zangle = PC.degtorad(zangle)

    return np.sqrt((r + y)**2 * np.cos(zangle)**2 + 2. * r * (1. - y) - y**2 + 1.0) - \
        (r + y) * np.cos(zangle)



import numpy as np


def foldAt(time, period, T0=0.0, sortphase=False, centralzero=False, getEpoch=False):
    """
    Fold time series with a particular period.

    Calculate the phase, P, from time, period, and reference point, T0,
    according to 

    .. math::
    
        P = (time - T0)/period - [(time-T0)/period].
    
    Here, square brackets indicate Gaussian brackets (i.e., the floor function), and
    the phase is a number between 0 and 1 by definition (and not between 0 and 2pi).
    
    Optionally, also the epoch, E, can be returned, which is an integer corresponding
    to the second term in the above equation. For any point of the series, therefore,
    the following relation applies
    
    .. math::
    
        time = T0 + (E+P) * period .

    Of course the series to be folded does not necessarily have to be a time
    series although this particular example guides the naming convention here.
    
    Parameters
    ----------
    time : array
        The time stamps.
    period : float
        The period to fold with (same units as time stamps).
    T0 : float
        Time reference point. The point T0 as well as all points T0+n*period with
        integer n are mapped to phase zero. Default is 0.0.
    sortphase : boolean, optional
        If True, return values will be sorted w.r.t. phase
    centralzero: boolean, optional
        If True, phase will be between -0.5 and +0.5 instead of 0 to 1. 
    getEpoch : boolean, optional
        If True, an array holding the epoch for every point
        in time will be returned; the default is False.
        Note that the first epoch, corresponding to times between T0 and T0+per, is 0.

    Returns
    -------
    Phases : array
        The (unsorted) phase array pertaining to the input time axis.
        Sorted if `sortphase` is set True.
    Epoch : array, optional
        An array holding the epoch for every given point in time.
        The counting starts at zero. Only returned if `getEpoch`
        is True.
    """
    epoch = np.floor((time - T0) / period)
    phase = (time - T0) / period - epoch
    
    if centralzero:
        indi = np.where(phase >= 0.5)[0]
        phase[indi] -= 1.0
    
    if sortphase:
        indi = np.argsort(phase)
        phase, epoch = phase[indi], epoch[indi]
    
    if getEpoch:
        return phase, epoch
    return phase



import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE

class BallesterosBV_T:
  """
    Black-body based conversion between effective temperature and B-V color.
    
    Ballesteros 2012 (EPL 97, 34008) present a conversion between
    effective temperature and B-V color index based on a black body
    spectrum and the filter functions.
  """
  
  def __init__(self):
    self._a = 0.92
    self._b = 1.7
    self._c = 0.62
    self._T0 = 4600.0
  
  def bv2T(self, bv):
    """
      Convert B-V color into temperature.
      
      Parameters
      ----------
      bv : float
          B-V color index [mag]
      
      Returns
      -------
      T : float
          Temperature [K]
    """
    T = self._T0*(1.0/(self._a*bv + self._b) + 1.0/(self._a*bv + self._c))
    return T
  
  def t2bv(self, T):
    """
      Convert temperature into B-V color.
      
      Parameters
      ----------
      T : float
          Temperature in K.
      
      Returns
      -------
      bv : float
          B-V color index [mag].
    """
    z = T/self._T0
    ap = z*self._a**2
    bp = self._a*self._c*z + self._b*self._a*z - 2.0*self._a
    cp = self._b*self._c*z -self._c - self._b
    
    sqrtarg = bp**2 - 4.0*ap*cp
    # By comparison with a BB it can be verified that
    # the physical solution is this one
    bv1 = (-bp + np.sqrt(sqrtarg))/(2.0*ap)
    return bv1
    
    
def bv2T_Ballesteros(bv):
  b = BallesterosBV_T()
  return b.bv2T(bv)
bv2T_Ballesteros.__doc__ = BallesterosBV_T.bv2T.__doc__


def t2bv_Ballesteros(T):
  b = BallesterosBV_T()
  return b.t2bv(T)
t2bv_Ballesteros.__doc__ = BallesterosBV_T.t2bv.__doc__


from __future__ import print_function, division
import unittest


class TestPyPDMExSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testsanity_Scanner(self):
        # Import PDM module
        from PyAstronomy.pyTiming import pyPDM

        # Get Scanner instance
        scanner = pyPDM.Scanner(minVal=0.5, maxVal=1.0, dVal=0.05, mode="period")
        # Print the periods covered by the scanner
        print("Periods: ", end=" ")
        for period in scanner:
            print(period, end=" ")

    def testsanity_PDMAna(self):
        import numpy
        import matplotlib.pylab as plt
        from PyAstronomy.pyTiming import pyPDM

        # Create artificial data with frequency = 3,
        # period = 1/3
        x = numpy.arange(100) / 100.0
        y = numpy.sin(x * 2.0 * numpy.pi * 3.0 + 1.7)

        # Get a ``scanner'', which defines the frequency interval to be checked.
        # Alternatively, also periods could be used instead of frequency.
        S = pyPDM.Scanner(minVal=0.5, maxVal=5.0, dVal=0.01, mode="frequency")

        # Carry out PDM analysis. Get frequency array
        # (f, note that it is frequency, because the scanner's
        # mode is ``frequency'') and associated Theta statistic (t).
        # Use 10 phase bins and 3 covers (= phase-shifted set of bins).
        P = pyPDM.PyPDM(x, y)
        f1, t1 = P.pdmEquiBinCover(10, 3, S)
        # For comparison, carry out PDM analysis using 10 bins equidistant
        # bins (no covers).
        f2, t2 = P.pdmEquiBin(10, S)

        # Show the result


#     plt.figure(facecolor='white')
#     plt.title("Result of PDM analysis")
#     plt.xlabel("Frequency")
#     plt.ylabel("Theta")
#     plt.plot(f1, t1, 'bp-')
#     plt.plot(f2, t2, 'gp-')
#     plt.legend(["pdmEquiBinCover", "pdmEquiBin"])
#     plt.show()



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import numpy as np
from .params import Params
import re
import copy
import types
from PyAstronomy.pyaC import pyaErrors as PE
from .nameIdentBase import ModelNameIdentBase
from PyAstronomy import pyaC
from time import time as timestamp
from .fufDS import FufDS
from .extFitter import NelderMead
import six
import six.moves as smo

from PyAstronomy.funcFit import _pymcImport, _scoImport, ic

if _pymcImport:
    import pymc
if _scoImport:
    import scipy.optimize as sco
if ic.check["emcee"]:
    import emcee
if ic.check["progressbar"]:
    import progressbar


def addEval(self, x):
    return (self.leftCompo.evaluate(x) + self.rightCompo.evaluate(x))


def subEval(self, x):
    return (self.leftCompo.evaluate(x) - self.rightCompo.evaluate(x))


def divEval(self, x):
    return (self.leftCompo.evaluate(x) / self.rightCompo.evaluate(x))


def mulEval(self, x):
    return (self.leftCompo.evaluate(x) * self.rightCompo.evaluate(x))


def powEval(self, x):
    return (self.leftCompo.evaluate(x) ** self.rightCompo.evaluate(x))


class MiniFunc:
    r"""
    This decorator can be applied to use
    self-defined objective functions.

    Applied to an objective function, it adds the functionality
    needed to evaluate the model given a certain parameter vector,
    so that the user does only have to take care
    about the quantity to be minimized.

    Parameters
    ----------
    odf : fitting object
        The fitting object that is supposed to use the self-defined
        objective function.
    """

    def __init__(self, odf):
        r"""
        Parameter:
          - `odf` - An instance of a fitting objects such as for example *GaussFit1d*.
        """
        # Save a REFERENCE to the fitting object
        self.odf = odf

    def __call__(self, f):
        r"""
        Parameter:
         - `f` - The user-defined objective function.
        """
        def miniFunc(P):
            # Update the parameter values in the 'Params' class instance.
            self.odf.pars.setFreeParams(P)
            # Obtain penalties
            pfac, pdict = self.odf.pars.getPenalty(
                penaltyFact=self.odf.penaltyFactor)
            # Assign penalty
            val = pfac
            # Apply conditional restrictions
            val += self.odf.pars.applyConditionalRestrictions()
            # Assign x, y, and yerr attributes. This is a not-so-beautiful
            # way of not breaking the API and having the funcFit data
            # storage object.
            self.odf.x = self.odf._fufDS.x
            self.odf.y = self.odf._fufDS.y
            self.odf.yerr = self.odf._fufDS.yerr

            try:
                # Update self.model to hold the evaluated function.
                self.odf.updateModel()
            except Exception as e:
                if val > 0.0:
                    # Allow model evaluation to fail if parameters wandered outside the
                    # valid range (returning immediately maybe an option, but this largely maintains
                    # behavior, preventing annoying errors)
                    return val
                # If no penalty is present, raise exception
                raise (PE.PyAAlgorithmFailure("Could not evaluate model for parameters: " + str(self.odf.parameters()),
                                              where="ondeDFit",
                                              tbfe=e,
                                              solution=["Try to define 'restrictions' via setRestriction if parameter values are invalid.",
                                                        "Adjust implementation of model to prevent error."]))
            # Add value of actual objective function
            val += f(self.odf, P)
            return val
        return miniFunc


class _PyMCSampler:
    r"""
    This class encapsulates a number of methods helping to set up the
    PyMC sampler.
    """

    def _dictComplete(self, l, d, whichDict, forget=[]):
        r"""
        Checks whether the list `l` contains all keys present in dictionary `d`.

        Parameters:
          `l` - A list of strings
          `d` - A dictionary with string keys.
          `whichDict` - string, Which dictionary (e.g., start values) is under consideration?
          `forget` - A list of string specifying keys, which can be omitted in d.
        """
        if len(l) != (len(d) + len(forget)):
            message = "Dictionary of " + whichDict + \
                " has not the correct number of entries - "
            if len(l) < (len(d) + len(forget)):
                message += " there are too many.\n"
            else:
                message += " there are too few.\n"
            message += "  Needed entries: " + ', '.join(l) + "\n"
            given = list(d)
            given.extend(forget)
            message += "  Given entries: " + ', '.join(given)
            raise (PE.PyAValError(message, where="fitMCMC",
                                  solution="Adjust input dictionary."))
        for elem in l:
            if elem in forget:
                continue
            if elem not in d:
                message = "Error in " + whichDict + " dictionary.\n"
                message += "  Key " + elem + " is missing!"
                raise (PE.PyAValError(message, where="fitMCMC",
                                      solution="Adjust input dictionary."))

    def _basicStatMCMCOutput(self, bs):
        print("Basic statistics of MCMC analysis: ")
        print("-----------------------------------------------------")
        for k in six.iterkeys(bs):
            print("Parameter: ", k)
            if bs[k] is None:
                print("  No output available!")
                continue
            for k2, v2 in six.iteritems(bs[k]):
                if k2 != "quantiles":
                    print("  " + k2 + " : " + str(v2))
                else:
                    s = "  quantiles : "
                    for qk, qv in six.iteritems(bs[k]["quantiles"]):
                        print(s + str(qk) + " : " + str(qv))
                        s = "              "
            print("-----------------------------------------------------")

    def _checkDbArgs(self, dbArgs):
        r"""
        Check whether database arguments are given. If not, use default parameters.
        """
        # If no db is specified, use pickle
        if not "db" in dbArgs:
            dbArgs["db"] = "pickle"

        if dbArgs["db"] == "pickle":
            if not "dbname" in dbArgs:
                dbArgs["dbname"] = "tmp.pickle"
            return dbArgs
        elif dbArgs["db"] == "hdf5":
            if not "dbname" in dbArgs:
                dbArgs["dbname"] = "tmp.zlib"
            if not "dbmode" in dbArgs:
                dbArgs["dbmode"] = "w"
            if not "dbcomplevel" in dbArgs:
                dbArgs["dbcomplevel"] = 9
            if not "dbcomplib" in dbArgs:
                dbArgs["dbcomplib"] = "zlib"
            return dbArgs
        else:
            raise (PE.PyAValError("Database: " + str(dbArgs["db"] + " currently not supported.",
                                                     solution="Use another database (e.g., pickle or hdf5).")))
        return dbArgs

    def MCMCautoParameters(self, ranges, picky=True, stepsize=1e-2, setRestrictionsFromPriors=False):
        r"""
        Convenience function to generate parameters for MCMC fit.

        This function prepares the `X0`, `Lims`, and `Steps` dictionaries
        needed to call :py:func:`fitMCMC`.

        For `X0`, the current parameter values are used. `Lims` is constructed
        using the `ranges` parameter, and `Steps` is defined on the basis
        of the `stepsize` and `ranges` parameters.

        The `picky` parameter determines how this functions handles behaves
        if it encounters parameters in the `range` dictionary, which were either
        not thawed or have been thawed but have not been specified in `ranges`.
        If `picky` is True (the default), this function throws an error if
        `ranges` does not cover all and only the free parameters. If picky is
        False, the function will automatically thaw all parameters specified
        through `ranges` and freeze the rest.

        .. warning:: There is NO guarantee that the sampling parameters
                     (start values, limits for the uniform priors, and initial
                     step sizes for the sampler) are reasonable. You need to
                     check the results.

        Parameters
        ----------
        ranges : dictionary
          Holds the fit ranges for the individual parameters. If single values 
          are given, the sampling range (uniform prior) will be arranged
          symmetrically around the parameter's current value.
          It is also possible to specify a range
          directly using, e.g., "A1":[0,100].
        stepsize : float, optional
          Defines the step size as a fraction of the fit range
          given in `ranges`.
        picky : boolean, optional
          If True (default), the list of free parameters has to match exactly
          the list of parameters specified in `ranges`. If False, the list
          of free parameters will be adapted to those given in `ranges`.
        setRestrictionsFromPriors : boolean, optional 
          Default: False. If True, parameter restrictions are applied according to
          the ranges of the uniform priors.

        Returns
        -------
        X0 : dictionary
            Maps parameter name to start value.
        lims : dictionary
            Maps parameter name to [lower, upper] limit.
        steps : dictionary
            Maps parameter name to step size. 

        Examples
        --------

        ::

          from PyAstronomy import funcFit as fuf
          import numpy as np
          import matplotlib.pylab as plt

          x = np.linspace(0,30,1000)
          gauss = fuf.GaussFit1d()
          gauss["A"] = 1
          gauss["mu"] = 23.
          gauss["sig"] = 0.5
          yerr = np.random.normal(0., 0.05, len(x))
          y = gauss.evaluate(x) + yerr
          # This step is not necessary if <picky>=False in MCMCautoParameters.
          gauss.thaw(["A","mu","sig"])
          X0, lims, steps = gauss.MCMCautoParameters({"A":[0,10],"mu":3, "sig":[0.1,1.0]})
          gauss.fitMCMC(x, y, X0, lims, steps, yerr=yerr, iter=1000)

          plt.plot(x, y, 'k+')
          plt.plot(x, gauss.evaluate(x), 'r--')
          plt.show()

        """

        X0 = {}
        steps = {}
        lims = {}

        # parameters which are free but not given
        missingParams = []
        # parameters which are not free but given
        extraParams = []

        # Check if ranges are given for all free parameters
        for p in self.freeParameters():
            if p not in ranges:
                if picky == False:
                    self.freeze(p)
                missingParams.append(p)
        # Check if all given ranges pertain to free parameters
        for p in ranges:
            if p not in self.freeParameters():
                if picky == False:
                    self.thaw(p)
                extraParams.append(p)
        # Raise exception if missing or extra parameters exits unless picky=False
        if picky:
            if len(missingParams) > 0:
                raise (PE.PyAValError("Not enough parameters: " + str(missingParams), where="_PyMCSampler::MCMCautoParameters",
                                      why="Ranges for these free fit parameters are missing."))
            if len(extraParams) > 0:
                raise (PE.PyAValError("Too many parameters: " + str(extraParams), where="_PyMCSampler::MCMCautoParameters",
                                      why="Ranges were given for non-free parameters."))

        if stepsize >= 1.0:
            raise (PE.PyAValError("The step size is larger than one. Must be a fraction of the range, i.e., smaller than 1.",
                                  solution="Redefine the step size. Use, e.g., 0.01."))

        # Loop through given parameters
        for p in ranges:
            X0[p] = self[p]
            # range is [a, b] or (a, b) then use this range, otherwise symmetric around
            #    the current value
            if isinstance(ranges[p], (list, tuple)):
                lims[p] = [ranges[p][0], ranges[p][1]]
            else:
                lims[p] = [X0[p] - ranges[p] / 2., X0[p] + ranges[p] / 2.]
            steps[p] = (max(lims[p]) - min(lims[p])) * stepsize
            if setRestrictionsFromPriors:
                self.setRestriction({p: lims[p]})

        return X0, lims, steps

    def autoFitMCMC(self, x, y, ranges, picky=True, stepsize=1e-3, yerr=None,
                    pymcPars={}, pyy=None, potentials=[], dbfile="mcmcSample.tmp",
                    dbArgs={}, **sampleArgs):
        r"""
        Convenience function to using auto-generated sampling parameters in MCMC.

        This function is essentially a wrapper around :py:func:`fitMCMC`.
        It allows you to use the 
        start values, step sizes, and limits constructed using
        :py:func:`MCMCautoParameters`.
        This method takes all parameters but `X0`, `Lims`, and `Steps`,
        which are expected by :py:func:`fitMCMC`. Additionally,
        the following parameters are available.

        .. warning:: There is NO guarantee that the sampling parameters
                     (start values, limits for the uniform priors, and initial
                     step sizes for the sampler) are reasonable. You need to
                     check the results.

        Parameters
        ----------
        ranges : dictionary
          Holds the fit ranges for the individual parameters. If a single values 
          is given, the fit range will be arranged symmetrically around the 
          parameter's current value. It is also possible to specify a range
          directly using, e.g., "A1":[0,100].
        stepsize : float, optional
          Defines the step size as a fraction of the fit range
          given in `ranges`.
        picky : boolean, optional
          If True (default), the list of free parameters has to match exactly
          the list of parameters specified in `ranges`. If False, the list
          of free parameters will be adapted to those given in `ranges`.

        Examples
        --------
        ::

          from PyAstronomy import funcFit as fuf
          import numpy as np
          import matplotlib.pylab as plt

          x = np.linspace(0,30,1000)
          gauss = fuf.GaussFit1d()
          gauss["A"] = 1
          gauss["mu"] = 23.
          gauss["sig"] = 0.5
          yerr = np.random.normal(0., 0.05, len(x))
          y = gauss.evaluate(x) + yerr
          ranges = {"A":[0,5],"mu":10, "sig":[0.1,1.0]}
          gauss.autoFitMCMC(x, y, ranges, yerr=yerr, iter=1000, picky=False)

          plt.plot(x, y, 'k+')
          plt.plot(x, gauss.evaluate(x), 'r--')
          plt.show()
        """
        X0, lims, steps = self.MCMCautoParameters(
            ranges, picky=picky, stepsize=stepsize)
        self.fitMCMC(x, y, X0, lims, steps, yerr=yerr, pymcPars=pymcPars, pyy=pyy,
                     potentials=potentials, dbfile=dbfile, dbArgs=dbArgs, **sampleArgs)

    def fitMCMC(self, x, y, X0, Lims, Steps, yerr=None, pymcPars={}, pyy=None,
                potentials=[], dbfile="mcmcSample.tmp", dbArgs={}, **sampleArgs):
        raise (PE.PyANotImplemented("This method has to be implemented"))


class _OndeDFitParBase:

    def __init__(self, parList, **kwargs):
        self.pars = Params(parList)
        # Set up the propMap (used to create a mapping between "property" (e.g., mu)
        # and variable name (e.g., "mu_1"))
        self.propMap = {}
        for p in parList:
            self.propMap[p] = p
        # Enable Root naming and component counters
        self.naming = ModelNameIdentBase(**kwargs)

    def __getitem__(self, specifier, **kwargs):
        name = self.naming.convertSpecifier(specifier)[0]
        if not name in self.propMap:
            if name in self.parameters():
                return self.pars.__getitem__(name)
            else:
                raise (PE.PyAValError("No such parameter: " + str(name), where="OneDFit::__getitem__",
                                      why="Did you perhaps misspell the parameter name?"))
        return self.pars.__getitem__(self.propMap[name])

    def __setitem__(self, specifier, value):
        name = self.naming.convertSpecifier(specifier)[0]
        if not name in self.propMap:
            if name in self.parameters():
                self.pars.assignValue({name: value})
                return
            raise (PE.PyAValError("No such parameter: " + str(name), where="OneDFit::__setitem__",
                                  why="Did you perhaps misspell the parameter name?"))
        self.pars.assignValue({self.propMap[name]: value})

    def hasVariable(self, specifier):
        r"""
        Determine whether the variable exists.

        Parameters
        ----------
        specifier : string or tuple
            Defines the name of the variable to be checked by
            string or specifier.

        Returns
        -------
        Flag : boolean
            True if the variable exists.
        """
        varName = self.naming.specifierToName(specifier)
        return self.pars.hasParam(varName)

    def assignValue(self, specval):
        r"""
        Assign new values to variables.

        Parameters
        ----------
        specval : dictionary
            Dictionary mapping variable names (wither given as string
            or specifier tuple) to value.
        """
        return self.pars.assignValue(self.naming.specifierToName(specval))

    def assignValues(self, specval):
        return self.assignValue(specval)
    assignValues.__doc__ = assignValue.__doc__

    def thaw(self, specifiers):
        r"""
        Consider variables fixed.

        Parameters
        ----------
        specifiers : list of strings or tuples
            The names of the variables to be fixed. Either given as
            string or specifier tuple.
        """
        self.pars.thaw(self.naming.specifierToName(specifiers))

    def freeze(self, specifiers):
        r"""
        Consider variables free to float.

        Parameters
        ----------
        specifiers : list of strings or tuples
            The names of the variables to be thawed. Either given as
            string or specifier tuple.
        """
        self.pars.freeze(self.naming.specifierToName(specifiers))

    def setRestriction(self, restricts):
        r"""
        Define restrictions.

        Parameters
        ----------
        restricts : dictionary
            A dictionary associating a variable (given as string or
            specifier tuple) with a restriction of the form:
            [lower, upper]. Use "None" where no restrictions shall
            be applied.
        """
        self.pars.setRestriction(self.naming.specifierToName(restricts))

    def untie(self, parName, forceFree=False):
        r"""
        Remove all relations of parameter parName, i.e., the parameter is not dependend 
        on other parameters. The parameter parName is set to "freeze".

        Parameters
        ----------
        parName : string
            The name of the dependent variable which should become "unrelated".
        forceFree : boolean
            Set parName to "free" instead of "frozen" if set to True.
        """
        self.pars.untie(self.naming.specifierToName(
            parName), forceFree=forceFree)

    def relate(self, dependentVar, independentVars, func=None, **kwargs):
        r"""
        Define a relation.

        Parameters
        ----------
        dependentVar : string or tuple
            The dependent variable given by string or specifier tuple.
        independentVars : string or list of strings
            The independent variables. You may also use specifier tuples
            to address the variables.
        func : callable
            The functional form of the relation. Must take the
            independent variables as arguments and returns the
            value of the dependent variable. If None is given,
            equality will be assumed.
        """
        self.pars.relate(self.naming.specifierToName(
            dependentVar), self.naming.specifierToName(independentVars), func, **kwargs)

    def getRelationsOf(self, specifier):
        r"""
        Return relations of a variable.

        Parameters
        ----------
        specifier : string or tuple
            Variable name or specifier tuple.

        Returns
        -------
        Relations : list of relations
            Those relations in which `specifier` is the dependent variable, i.e.,
            specifier = f(x,y).
        """
        return self.pars.getRelationsOf(self.naming.specifierToName(specifier))

    def freeParameters(self):
        return self.pars.freeParameters()
    freeParameters.__doc__ = Params.freeParameters.__doc__

    def frozenParameters(self):
        return self.pars.frozenParameters()
    frozenParameters.__doc__ = Params.frozenParameters.__doc__

    def freeParamNames(self):
        return self.pars.freeParamNames()
    freeParamNames.__doc__ = Params.freeParamNames.__doc__

    def numberOfFreeParams(self):
        return self.pars.numberOfFreeParams()
    numberOfFreeParams.__doc__ = Params.numberOfFreeParams.__doc__

    def getRestrictions(self):
        return self.pars.getRestrictions()
    getRestrictions.__doc__ = Params.getRestrictions.__doc__

    def delRestriction(self, parName):
        return self.pars.delRestriction(parName)
    delRestriction.__doc__ = Params.delRestriction.__doc__

    def availableParameters(self):
        return self.pars.availableParameters()
    availableParameters.__doc__ = Params.availableParameters.__doc__

    def parameters(self):
        return self.pars.parameters()
    parameters.__doc__ = Params.parameters.__doc__

    def saveState(self, *args, **kwargs):
        return self.pars.saveState(*args, **kwargs)
    saveState.__doc__ = Params.saveState.__doc__

    def restoreState(self, resource):
        self.pars.restoreState(resource)
    restoreState.__doc__ = Params.restoreState.__doc__

    def addConditionalRestriction(self, *args):
        return self.pars.addConditionalRestriction(*args)
    addConditionalRestriction.__doc__ = Params.addConditionalRestriction.__doc__

    def removeConditionalRestriction(self, *args):
        return self.pars.removeConditionalRestriction(*args)
    removeConditionalRestriction.__doc__ = Params.removeConditionalRestriction.__doc__

    def showConditionalRestrictions(self, **kwargs):
        return self.pars.showConditionalRestrictions(**kwargs)
    showConditionalRestrictions.__doc__ = Params.showConditionalRestrictions.__doc__


class IFitterBase:
    r"""
    Base class for internal fitter.
    """

    def __init__(self):
        self._objfval = None

    def fit(self, minifunc, x0):
        r"""
        Carry out the minimization.

        Parameters
        ----------
        minifunc : callable
            Objective function.
        x0 : list
            Starting values

        Returns
        -------
        Result : list or tuple
            First item is a list of the best-fit values and second
            item is the value of the objective function.
        """
        raise (PE.PyANotImplemented(
            "IFitterBase: fit method needs to be implemented"))

    def _digestkwargs(self, kwargs):
        r"""
        Check validity of keywords.

        Tests whether all given keywords are included in the list
        of "allowed" keywords (attribute _allowedKWs).

        Parameters
        ----------
        kwargs : dictionary
            Keyword arguments handed to the minimizer.
        """
        if not hasattr(self, "_allowedKWs"):
            raise (PE.PyANotImplemented("No _allowedKWs attribute. Must be set to use '_digestkwargs'.",
                                        solution="Specify _allowedKWs, e.g., in constructor."))
        for k in six.iterkeys(kwargs):
            if not k in self._allowedKWs:
                raise (PE.PyAValError("Keyword " + k + " not allowed in algorithm: " + self.name,
                                      solution="Allowed keywords: " +
                                      ', '.join(self._allowedKWs),
                                      where=self.name))

    def getObjFuncValue(self):
        r"""
        Access value of objective function.

        Returns
        -------
        value : float
            Value of objective function.
        """
        return self._objfval

    def __call__(self, *args, **kwargs):
        r"""
        Wrapper around the actual fit method.
        """
        return self.fit(*args, **kwargs)


class ScipyFMIN(IFitterBase):
    r"""
    Wrapper around scipy.optimize.fmin.
    """

    def __init__(self, *args, **kwargs):
        IFitterBase.__init__(self)
        self._allowedKWs = ["xtol", "ftol", "maxiter",
                            "maxfun", "disp", "retall", "callback"]
        self.name = "scipy.optimize.fmin"

    def fit(self, miniFunc, x0, *fminpars, **fminargs):
        r"""
        Wrapper around scipy.optimize.fmin.
        """
        self._digestkwargs(fminargs)
        self._result = sco.fmin(miniFunc, x0, *fminpars,
                                full_output=True, **fminargs)
        self._objfval = self._result[1]
        return self._result[0], self._result[1]


class ScipyFMIN_POWELL(IFitterBase):
    r"""
    Wrapper around scipy.optimize.fmin_powell.
    """

    def __init__(self, *args, **kwargs):
        IFitterBase.__init__(self)
        self._allowedKWs = ["xtol", "ftol", "maxiter",
                            "maxfun", "disp", "retall", "callback", "direc"]
        self.name = "scipy.optimize.fmin_powell"

    def fit(self, miniFunc, x0, *fminpars, **fminargs):
        r"""
        Wrapper around scipy.optimize.fmin_powell
        """
        self._digestkwargs(fminargs)
        self._result = sco.fmin_powell(miniFunc, x0, *fminpars,
                                       full_output=True, **fminargs)
        self._objfval = self._result[1]
        return self._result[0], self._result[1]


class FuFNM(IFitterBase):

    def __init__(self, *args, **kwargs):
        IFitterBase.__init__(self)
        self._allowedKWs = ["initDelta", "maxIter", "callback", "nmCritLim"]
        self.name = "funcFit NM65"
        self._obj = args[0]
        self._nm = NelderMead(**kwargs)

    def fit(self, miniFunc, x0, *fminpars, **fminargs):
        r"""
        Wrapper around funcFit's NelderMead implementation.

        See the implementation of the `fit` method of :py:func:`NelderMead`
        for the available keyword arguments (fminargs).
        """
        self._digestkwargs(fminargs)
        self._bestFit = self._nm.fit(
            self._obj, self._obj._fufDS, self._obj.miniFunc, **fminargs)
        self._bestFitVals = self._obj.pars.getFreeParams()
        self._objfval = self._obj.miniFunc(self._bestFitVals)
        return self._bestFitVals, self._objfval


FuFNM.__doc__ = NelderMead.__doc__


class FuFPrior(object):
    r"""
    A number of priors.

    Properly initialized, an object of type "FuFPrior"
    is callable. On call, it expects a dictionary with
    the names and values of the parameters as first
    argument and the name of the parameter under consideration
    as second argument. The return value should be the
    (natural) logarithm of the associated prior probability
    distribution.

    Parameters
    ----------
    lnp : string, {uniform, jeffreyPS, gaussian, limuniform}
        uniform : improper uniform prior. limuniform: proper
        uniform prior. 'lower' and 'upper' define the lower and
        upper bounds of the interval.
        jeffreyPS: Jeffreys
        prior for a Poisson scaling parameter. gaussian: A
        Gaussian prior. The keyswords 'mu' and 'sig' must be
        specified to define the mean and standard deviation of
        the Gaussian.
    """

    def _uniform(self, **kwargs):
        def uniform(ps, n, **rest):
            return 0.0
        return uniform

    def _uniformLimit(self, **kwargs):
        if kwargs["upper"] < kwargs["lower"]:
            raise (PE.PyAValError("upper needs to be larger than lower",
                                  where="FuFPrior (limited uniform distribution)",
                                  solution="Adapt upper and lower."))
        p = np.log(1.0 / (kwargs["upper"] - kwargs["lower"]))

        def unilimit(ps, n, **rest):
            if (ps[n] >= kwargs["lower"]) and (ps[n] <= kwargs["upper"]):
                return p
            else:
                return -np.inf
        return unilimit

    def _jeffreyPoissonScale(self, **kwargs):
        def jps(ps, n, **rest):
            return -0.5 * np.log(ps[n])
        return jps

    def _gaussian(self, **kwargs):
        r = -0.5 * np.log(2.0 * np.pi * kwargs["sig"]**2)

        def gaussianPrior(ps, n, **rest):
            return r - (ps[n] - kwargs["mu"])**2 / (2.0 * kwargs["sig"]**2)
        return gaussianPrior

    def _callDelegator(self, *args, **kwargs):
        """ Overwritten by the method to represent __call__ """
        raise (PE.PyANotImplemented("_callDelegator is not implemented."))

    def __call__(self, *args, **kwargs):
        return self._callDelegator(*args, **kwargs)

    def __init__(self, lnp, **kwargs):
        if isinstance(lnp, six.string_types):
            if lnp == "uniform":
                self._callDelegator = self._uniform(**kwargs)
            elif lnp == "limuniform":
                self._callDelegator = self._uniformLimit(**kwargs)
            elif lnp == "jeffreyPS":
                self._callDelegator = self._jeffreyPoissonScale(**kwargs)
            elif lnp == "gaussian":
                self._callDelegator = self._gaussian(**kwargs)
            else:
                raise (PE.PyAValError("No prior defined for " + str(lnp),
                                      where="FuFPrior",
                                      solution="Use either of {uniform, limuniform, jeffreyPS, gaussian}"))


class OneDFit(_OndeDFitParBase, _PyMCSampler):
    r"""
    The base class for fitting objects.

    Parameters
    ----------
    parList : list of strings
        Contains the names of the properties
        defining the model. By default, variables of the same name
        are used to represent them.

    Attributes
    ----------
    model : array
        Used by the `updateModel` method to store the
        evaluated model for current parameter settings.
        Holds the best-fit model after a call to a fit method.
    penaltyFactor : float
        The penalty factor used to apply penalties for
        enforcing restrictions (default = 10**20).

    Notes
    -----

    The purpose of the class

      The purpose of this class is to provide a convenient interface
      to various fitting algorithms.
      It provides the functionality, which allows for parameter
      fitting, but does not implement a particular model.
      The class can be used to fit any kind of model, which has
      to be implemented in a class, which inherits from
      the *OneDFit* class.

    Management of fitting parameters

      The fitting parameters are managed by a *Params* class
      instance, which provides a wealth of possibilities to
      influence the behavior of the parameters during the
      fitting process. This includes deciding whether a particular
      parameter is supposed to be a free fitting parameter,
      applying restrictions to limit the valid range for a
      parameters, or the introduction of functional dependencies among
      different parameters.

    Properties versus variable names

      Each model is described by a number of *properties*, such
      as, for example, mass and radius. These may be represented
      by arbitrarily named variables. Normally, it is convenient to
      name the variables according to the properties they describe,
      which is the default behavior. However, in some cases, for example
      if a model consists of two equal subcomponents, such a naming
      scheme leads to nonunique variable names, which has to be avoided.
      Now it is necessary to distinguish between the *property* and
      the describing variable. This class uses the `propMap`
      dictionary, which maps property name to
      variable name to manage these situations.

    Combining fitting objects

      Often, it can be convenient to combine a number of simple models
      to form a new, more complex one. The *OneDFit* class allows to
      combine objects using the arithmetic operators +-\*/,
      and the power (\*\*) operator.

    Naming scheme for models

      For simple models it is convenient to use a one-to-one mapping
      between property and variable name. It may, however, become
      necessary to deviate from this scheme, for example, to keep
      variable names unique. This class supports the following naming scheme:
      Each model has a "root name", which is supposed to be a concise
      string describing the model (for instance, "Gaussian"). The root
      name is attached to the property name using an underscore.
      If a complex model consists of more than one component
      with the same root name, a component counter, enclosed in
      parenthesis, is attached to the variable name.
      A variable name could, for example, look like: "mu_Gaussian(1)".

    Methods to be implemented in a model class

      A valid model class inheriting this interface class must provide
      the following methods

     - **__init__()** - The constructor.

                        Defines the set of properties describing the model.
     - **evaluate(x)** - An *evaluate* method.

                         This method takes a single
                         argument, x, which is an array of points at which
                         the model is to be evaluated. To access the
                         current model parameters, this method should use
                         the set/getitem methods. The return value is
                         an array holding the model evaluated at the
                         points given by `x`.
    """

    def __init__(self, parList, **kwargs):
        _OndeDFitParBase.__init__(self, parList, **kwargs)
        # Left and right compo(nent) are necessary for combining models
        self.leftCompo = None
        self.rightCompo = None
        self.penaltyFactor = 1e20
        self.model = None
        self._fufDS = None
        self.fitResult = None
        # Determines whether steppar can be used
        self._stepparEnabled = False

    def _compoWalk(self):
        r"""
        TBD
        """
        def walk(c, refs):
            refs.append(c)
            if c.leftCompo is not None:
                walk(c.leftCompo, refs)
            if c.rightCompo is not None:
                walk(c.rightCompo, refs)

        refs = []
        walk(self, refs)
        for c in refs:
            yield c

    def renameVariable(self, oldName, newName):
        r"""
        Change name of variable.

        Parameters
        ----------
        oldName : string
            Current variable name.
        newName : string
            New variable name.

        Notes
        -----
        Variable names and properties are not the same.
        """
        # First, walk down the left and right components (for combined models)
        # and change the variable names.
        if self.leftCompo is not None:
            try:
                self.leftCompo.renameVariable(oldName, newName)
            except PE.PyAValError:
                pass
        if self.rightCompo is not None:
            try:
                self.rightCompo.renameVariable(oldName, newName)
            except PE.PyAValError:
                pass
        # Now do the same for the "top" component
        if newName == oldName:
            # Ignore identical transformations
            return
        if newName in list(self.propMap.values()):
            raise (PE.PyANameClash("A variable named " + newName +
                                   " does already exist.", where="OneDFit::renameVariable"))
        if newName in self.propMap:
            if self.propMap[newName] != oldName:
                raise (PE.PyANameClash("You may not assign a name to a variable, which corresponds to the name of another property.",
                                       where="OneDFit::renameVariable"))
        if not oldName in list(self.propMap.values()):
            raise (PE.PyAValError("A variable named " + oldName +
                                  " does not exist.", where="OneDFit::renameVariable"))
        for k in six.iterkeys(self.propMap):
            if self.propMap[k] == oldName:
                self.propMap[k] = newName
                break
        # Tell the parameter class about the renaming (if this did not already happen)
        if not newName in six.iterkeys(self.pars.parameters()):
            self.pars.renameParameter(oldName, newName)

    def _isComposed(self):
        r"""
        Determines whether current model is "composed".

        A model is composed, if there are left and right components.

        Returns True if model is composed and False otherwise.
        """
        return ((self.leftCompo is not None) and (self.rightCompo is not None))

    def __combineRemapping(self, left, right):
        r"""
        This member is essentially a renaming machine. When combining models
        it can easily happen that two variables share the same name. If the
        models are combined, unique variable names are needed. This method
        uses the "root name" and "component counter" to assign new, unique
        names to the variables. 

        Parameters:
          - `left`, `right` - Two fitting objects (derived from OneDFit).
        """

        def extendCoDat(coDat, c):
            ident = c.naming.getRoot()
            if not ident in coDat:
                coDat[ident] = [c]
            else:
                coDat[ident].append(c)
            return coDat

        # Build up a dictionary assigning root to a list of corresponding components
        coDat = {}
        for c in left._compoWalk():
            extendCoDat(coDat, c)
        for c in right._compoWalk():
            extendCoDat(coDat, c)

        for k in six.iterkeys(coDat):
            # Loop over all available root names
            if len(coDat[k]) == 1:
                # Only a single component with that root, no problem, no renaming
                continue
            elif len(coDat[k]) >= 2:
                # Get all component counters pertaining to that root
                cs = np.array([c.naming.getComponentCounter()
                               for c in coDat[k]])
                # What is the current maximum
                highest = np.max(cs)
                # Check whether one or more counters are zero;
                # if so, increase the number and set it to "highest + 1".
                for c in coDat[k]:
                    if c.naming.getComponentCounter() == 0:
                        c.naming.setComponentCounter(highest + 1)
                        highest += 1
                # Now re-check whether counters are unique
                while True:
                    cs = np.array([c.naming.getComponentCounter()
                                   for c in coDat[k]])
                    u = np.unique(cs)
                    if len(u) == len(cs):
                        # All counters are unique, work is finished
                        break
                    # Set lowest counter
                    again = False
                    for c in coDat[k]:
                        # Loop over components
                        if len(np.where(cs == c.naming.getComponentCounter())[0]) > 1:
                            # If there are nonunique counters, make them unique and retry
                            c.naming.setComponentCounter(highest + 1)
                            highest += 1
                            again = True
                            break
                    if not again:
                        break

        def renameIfNecessary(c, baseComponent):
            # Rename all variables according to standard naming rules
            for prop in six.iterkeys(c.propMap):
                # Compose new name using "standard conventions"
                newName = c.naming.composeVariableName(prop)
                if not (newName == c.propMap[prop]):
                    # Name has changed
                    baseComponent.renameVariable(c.propMap[prop], newName)

        # Rename all variables mentioned in NONCOMPOSED models
        for c in left._compoWalk():
            if c._isComposed():
                continue
            renameIfNecessary(c, left)
        for c in right._compoWalk():
            if c._isComposed():
                continue
            renameIfNecessary(c, right)

    def _newPars(self, npars):
        r"""
        Change pars reference for left/right Compo recursively.
        """
        if self.leftCompo is not None:
            self.leftCompo.pars = npars
            self.leftCompo._newPars(npars)
        if self.rightCompo is not None:
            self.rightCompo.pars = npars
            self.rightCompo._newPars(npars)

    def __combineFittingObjects(self, right):
        r"""
        Creates and returns a fitting object combining the properties/variables of
        the current objects and that given by `right`.

        Parameters:
          - `right` - A fitting object (derived from OneDFit).
        """
        # Obtain deep copies of the involved fitting objects
        # The actual reason for doing this, is that the evaluate
        # methods have to be conserved (copied).
        left = copy.deepcopy(self)
        right = copy.deepcopy(right)
        # Find new parameter names to avoid parameter clashes
        # (parameter names must be unique)
        self.__combineRemapping(left, right)
        # Combine the Param instances
        npars = left.pars + right.pars
        # Obtain a new Fitting object
        result = OneDFit(list(npars.parameters().keys()))
        # Assign new combined Param instance
        result.pars = npars
        # Assign new parameter class to ``old'' individual fitting
        # objects.
        left.pars = result.pars
        right.pars = result.pars

        # Assign left and right component (for evaluation)
        result.leftCompo = left
        result.rightCompo = right
        # Recursively replace pars property by new pars reference
        result._newPars(npars)
        result.setRootName("combinedModel")
        return result

    def __add__(self, right):
        result = self.__combineFittingObjects(right)
        # For a discussion about bound instance methods see
        # https://stackoverflow.com/questions/972/adding-a-method-to-an-existing-object-instance
        result.evaluate = types.MethodType(addEval, result)
        # Save the 'operator' relating left and right component
        result._operator = '+'
        return result

    def __sub__(self, right):
        result = self.__combineFittingObjects(right)
        result.evaluate = types.MethodType(subEval, result)
        # Save the 'operator' relating left and right component
        result._operator = '-'
        return result

    def __mul__(self, right):
        result = self.__combineFittingObjects(right)
        result.evaluate = types.MethodType(mulEval, result)
        # Save the 'operator' relating left and right component
        result._operator = '*'
        return result

    def __div__(self, right):
        result = self.__combineFittingObjects(right)
        result.evaluate = types.MethodType(divEval, result)
        # Save the 'operator' relating left and right component
        result._operator = '/'
        return result

    def __pow__(self, right):
        result = self.__combineFittingObjects(right)
        result.evaluate = types.MethodType(powEval, result)
        # Save the 'operator' relating left and right component
        result._operator = '**'
        return result

    def __sqrDiff(self):
        @MiniFunc(self)
        def miniSqrDiff(odf, P):
            # Calculate squared difference
            chi = np.sum((self._fufDS.y - self.model)**2)
            return chi
        return miniSqrDiff

    def __chiSqr(self):
        @MiniFunc(self)
        def miniChiSqr(odf, P):
            # Calculate chi^2 and apply penalty if boundaries are violated.
            chi = np.sum(((self._fufDS.y - self.model) / self._fufDS.yerr)**2)
            return chi
        return miniChiSqr

    def __cash79(self):
        @MiniFunc(self)
        def miniCash79(odf, P):
            # Calculate Cash statistics according to Cash 1979 (ApJ 228, 939)
            c = -2.0 * np.sum(self._fufDS.y * np.log(self.model) - self.model)
            return c
        return miniCash79

    def setRootName(self, root, rename=False):
        r"""
        Define the root name of the model.

        Parameters
        ----------
        root : string
            A concise description of the model.
        rename : bool, optional, default=False
            If true, all model variables will be renaming using the root.
        """
        if not hasattr(self, "naming"):
            raise (PE.PyAUnclassifiedError("No attribute 'naming' found.",
                                           solution="Did you call the constructor of OneDFit before you tried to access this? If not, you should do that."))
        self.naming.setRootName(root)
        if rename:
            for par in six.iterkeys(self.propMap):
                oldName = self.propMap[par]
                newName = self.naming.composeVariableName(par)
                self.renameVariable(oldName, newName)

    def updateModel(self):
        r"""
        Recalculate the model using current settings.

        Notes
        -----
        Updates the `model` attribute of the class by
        calling `evaluate` using the `x` attribute, which
        is, e.g., assigned on call to a fit method.
        """
        if self._fufDS is None:
            return
        self.model = self.evaluate(self._fufDS.x)

    def setPenaltyFactor(self, penalFac):
        r"""
        Change the penalty factor.

        Parameters
        ----------
        penalFac : float
            The penalty factor (default is 1e20).

        Notes
        -----
        May also be done by accessing the `penalty` property
        directly.
        """
        self.penaltyFactor = penalFac

    def description(self, parenthesis=False):
        r"""
        Returns a description of the model based on the names of the individual components.

        Parameters
        ----------
        parenthesis : boolean, optional
            If True, the entire expression/description will be enclosed in
            parenthesis. The default is False.

        Returns
        -------
        Description : string
            Description of the model.
        """
        c = self
        if c._isComposed():
            if parenthesis:
                return "(" + c.leftCompo.description(True) + " " + c._operator + " " + c.rightCompo.description(True) + ")"
            else:
                return c.leftCompo.description(True) + " " + c._operator + " " + c.rightCompo.description(True)
        else:
            if c.naming.getComponentCounter() > 0:
                return c.naming.getRoot() + "(No. " + str(c.naming.getComponentCounter()) + ")"
            else:
                return c.naming.getRoot()

    def parameterSummary(self, toScreen=True, prefix="", sorting='none'):
        r"""
        Writes a summary of the parameters in text form.

        Parameters
        ----------
        toScreen : bool, optional, default = True
            If True, the output is written to the screen.
        prefix : string, optional, default = ""
            A prefix for each line (e.g., '#').
        sorting : string, optional, {'none', 'ps'}
            Determines the order in which the parameters are
            printed out. If 'none' is given (default), no
            particular order is imposed. If 'ps', Python's
            `sorting` routine is used to impose an order.

        Returns
        -------
        A list of strings containing the text.
        """
        lines = []

        # Collect roots and associated counters and
        # (root, counter) and associated component
        rootNums = {}
        rootNoComp = {}
        for c in self._compoWalk():
            if c._isComposed():
                continue
            # It is a noncomposed component
            if not c.naming.getRoot() in rootNums:
                rootNums[c.naming.getRoot()] = [c.naming.getComponentCounter()]
            else:
                rootNums[c.naming.getRoot()].append(
                    c.naming.getComponentCounter())
            rootNoComp[(c.naming.getRoot(), c.naming.getComponentCounter())] = c

        # Find maximal length of property, root, identifier (= root (counter)),
        # and variable. This is done to provide well formatted output
        maxPropLen = 0
        maxRootLen = 0
        maxIdentLen = 0
        maxVarLen = 0
        for k, c in six.iteritems(rootNoComp):
            maxRootLen = max(maxRootLen, len(k[0]))
            maxIdentLen = max(maxIdentLen, len(c.naming.identifier()))
            for prop, var in six.iteritems(c.propMap):
                maxPropLen = max(maxPropLen, len(prop))
                maxVarLen = max(maxVarLen, len(var))

        # Walk through all root and counter to provide output
        for root in sorted(rootNums.keys()):
            for counter in sorted(rootNums[root]):
                c = rootNoComp[(root, counter)]
                cono = "Parameters for Component: "
                if root == "":
                    cono += "unnamed"
                else:
                    cono += root
                if counter != 0:
                    cono += " (No. " + str(counter) + ")"
                lines.append("-" * len(cono))
                lines.append(cono)
                lines.append("-" * len(cono))

                # Sorting of the parameter names (component-wise)
                cprops = list(c.propMap.keys())
                if sorting == "none":
                    pass
                elif sorting == "ps":
                    cprops = sorted(cprops)
                else:
                    raise (PE.PyAValError("Unknown sorting mode: " + str(sorting),
                                          where="OneDFit::parameterSummary",
                                          solution="Use 'None' or 'ps'"))

                for prop in cprops:
                    var = c.propMap[prop]
                    # Loop over properties and variables
                    restricted = (self.getRestrictions()[var][0] is not None) or (
                        self.getRestrictions()[var][1] is not None)
                    lines.append(("Parameter: %" + str(maxPropLen) + "s  %" + str(maxIdentLen) + "s, [%" + str(maxVarLen) + "s], value: % 12g, free: %5s, restricted: %5s, related: %5s") %
                                 (prop, c.naming.identifier(), var, self[var], var in self.freeParameters(), restricted, (len(self.getRelationsOf(var)) != 0)))
                    app = " " * 4
                    if restricted:
                        res = self.getRestrictions()[var]
                        app += "Restriction: ["
                        if res[0] is not None:
                            app += "% g, " % res[0]
                        else:
                            app += "None, "
                        if res[1] is not None:
                            app += "% g]" % res[1]
                        else:
                            app += " None]"
                    relas = self.getRelationsOf(var)
                    if len(relas) != 0:
                        if restricted:
                            app += ", "
                        for r in relas:
                            app += " Relation: " + r[0] + " = f("
                            for k in range(len(r[2]) - 1):
                                app += r[2][k] + ", "
                            app += r[2][len(r[2]) - 1] + ")"

                    if re.match(r"\s*[^\s]+", app):
                        lines.append(app)

        for i in smo.range(len(lines)):
            lines[i] = prefix + lines[i]

        if toScreen:
            for l in lines:
                print(l)

        return lines

    def setObjectiveFunction(self, miniFunc="chisqr"):
        r"""
        Define the objective function.

        This function sets the `miniFunc` attribute, which is used
        to calculate the quantity to be minimized.

        Parameters
        ----------
        miniFunc : str {chisqr, cash79, sqrdiff} or callable
            The objective function. If "chisqr", chi-square will be
            minimzed. If "cash 79", the Cash statistics 
            (Cash 1979, ApJ 228, 939, Eq. 5) will be used.
            If "sqrdiff" is specified, 
            Otherwise, a user-defined function is assumed.
        """
        # Determine function to be minimized
        if miniFunc == "chisqr":
            self.miniFunc = self.__chiSqr()
            return
        elif miniFunc == "cash79":
            self.miniFunc = self.__cash79()
            return
        elif miniFunc == "sqrdiff":
            self.miniFunc = self.__sqrDiff()
            return
        else:
            if not hasattr(miniFunc, '__call__'):
                raise (PE.PyAValError("`miniFunc` is neither None, a valid string, or a function.",
                                      where="OneDFit::fit",
                                      solution="Use, e.g., 'chisqr' or another valid choice from the documentation."))

            # A function has been specified
            self.miniFunc = miniFunc
            return

    def fitMCMC(self, x, y, X0, Lims, Steps, yerr=None, pymcPars=None, pyy=None,
                potentials=None, dbfile="mcmcSample.tmp", quiet=False, dbArgs=None,
                adaptiveMetropolis=False, **sampleArgs):
        r"""
        Carry out MCMC fit/error estimation.

        This member is designed to provide a flexible but *easy to use*
        interface to the capabilities of pymc. In the simplest case,
        it assumes a Poisson or Gaussian distribution of data points
        and uses continuous, uniform variables (all free fitting
        variables) with starting values defined by `X0`, Limits
        given by `Lims`, and step sizes given by `Steps`
        to sample from the posterior.

        .. note::
           The result (the Marchov-Chain/PyMC MCMC-object) will
           be contained in the self.MCMC property; the output dictionary
           of MCMC.stats()
           (Mean, HPD interval etc.) is saved to self.basicStats.

        Parameters
        ----------
        x, y : array 
            Array providing the abscissa and ordinate values of the
            data points.
            For the MCMC process, it is essential to know
            the underlying distribution of the data points.
            `fitMCMC` assumes Poisson distributed data if
            `yerr` is not specified and Gaussian data if it
            is specified. If other distributions shall be used,
            the `pyy` parameter must contain a
            *PyMC* random variable specifying it.
        yerr : array, optional
            The error of the data points. If not specified, Poisson
            distributed data will be assumed.
        pyy : PyMC stochastic, optional,
            A PyMC variable containing the \
            data. This can be useful if the distribution is
            neither Poisson nor
            Gaussian as otherwise assumed by this method.
        X0 : dict
             A dictionary holding {"parName":value, ...} specifying
             the start values. Note that parameters treated in
             `pymcPars` must not be part of this.
        Lims : dict,
            A dictionary of type {"ParName:[min,max], ...}
            specifying the lower and upper limit of a variable.
            Note that parameters treated in pymcPars must not
            be part of this.
        Steps : dict
            A dictionary providing the step sizes for the MCMC sampler.
        pymcPars : dict, optional,
            Holds a dictionary of the form {"parName":PyMC-Variable, ...}.
            `pymcPars` can be used to specify a nonuniform
            distribution for a parameter.
        potentials : list of PyMC potentials, optional,
            Can be used to provide a list of PyMC potentials,
            which may be needed to provide priors.
        dbfile : string, optional,
            The name of the output file, which is to hold the MCMC chain
            produced during sampling (default="mcmcSample.tmp").
        quiet : bool, optional,
            Set to True in order to suppress the text output.
        sampleArgs : dict, optional,
            Here additional keywords can be specified, which
            will be handed to the `isample` member of PyMC. Most notably
            it is useful to specify **iter**, **burn**, and **thin**.
            For other possibilities see PyMC documentation.
        dbArgs : dict, optional,
            Keyword arguments specifying the trace database to be used.
            By default a pickle database named according to the
            `dbfile` keyword.
            You may also use an hdf5 database.
        adaptiveMetropolis : bool, optional,
            Set to true in order to use the AdaptiveMetropolis step method. 
        """
        global _pymcImport
        if not _pymcImport:
            raise (PE.PyARequiredImport("pymc package could not be imported.",
                                        solution="Install pymc (see http://code.google.com/p/pymc/"))
        if (pymc.__version__).startswith('3'):
            raise (PE.PyARequiredImport("pymc is available in version '" + str(pymc.__version__) + "'. " +
                                        "Please note that funcFit does not yet support pymc3, which is a " +
                                        "complete rewrite of pymc2.",
                                        solution="Please install pymc in version 2.3."))
        # Assign mutable default parameters
        if pymcPars is None:
            pymcPars = {}
        if dbArgs is None:
            dbArgs = {}
        if potentials is None:
            potentials = []
        # Assign attributes and check x, y, and yerr.
        self._fufDS = FufDS(x, y, yerr)
        # Copy the pymcPars and dbArgs dictionaries (prevents error on multiple sampler calls)
        pymcPars = pymcPars.copy()
        dbArgs = dbArgs.copy()
        # Get the names of the free parameters
        freeNames = self.freeParamNames()
        if not quiet:
            print("Free parameters: ", freeNames)
        # Check whether parameter lists are complete, define default steps
        # if necessary.
        self._dictComplete(freeNames, X0, "start values",
                           forget=list(pymcPars.keys()))
        self._dictComplete(freeNames, Lims, "limits",
                           forget=list(pymcPars.keys()))
        self._dictComplete(freeNames, Steps, "steps")

        # Define (or complete) the pymcPars dictionary by defining uniformly distributed
        # variables in the range [lim[0], lim[1]] with starting values defined by X0.
        for par in freeNames:
            if par in pymcPars:
                continue
            if not quiet:
                print("Using uniform distribution for parameter: ", par)
                print("  Start value: ",
                      X0[par], ", Limits = [", Lims[par][0], ", ", Lims[par][1], "]")
            pymcPars[par] = pymc.Uniform(
                par, lower=Lims[par][0], upper=Lims[par][1], value=X0[par], doc="Automatically assigned parameter.")

        # This function is used to update the model
        def getModel(**vals):
            self.assignValue(vals)
            self.updateModel()
            return self.model

        modelDet = pymc.Deterministic(
            eval=getModel,
            name='Model',
            parents=pymcPars,
            doc='The model',
            trace=True,
            verbose=0,
            dtype=float,
            plot=False,
            cache_depth=2)

        # Define the 'data' (y-values)
        if pyy is None:
            if yerr is None:
                if not quiet:
                    print(
                        "Assuming Poisson distribution for 'y'. Use 'pyy' parameter to change this!")
                pyy = pymc.Poisson(
                    "y", mu=modelDet, value=self._fufDS.y, observed=True)
            else:
                if not quiet:
                    print(
                        "Assuming Gaussian distribution for 'y'. Use 'pyy' parameter to change this!")
                pyy = pymc.Normal(
                    "y", mu=modelDet, tau=1.0 / self._fufDS.yerr**2, value=self._fufDS.y, observed=True)

        # Add data to the Model
        Model = [pyy]
        # Add potentials (e.g., priors)
        Model.extend(potentials)
        # Add free parameters
        for v in six.itervalues(pymcPars):
            Model.append(v)

        # Check database arguments
        if not "dbname" in dbArgs:
            dbArgs["dbname"] = dbfile
        dbArgs = self._checkDbArgs(dbArgs)

        if not quiet:
            print("Using database arguments: ", dbArgs)
        self.MCMC = pymc.MCMC(Model, **dbArgs)

        # Tell the MCMC class to use the MH algorithm with specified step width
        if adaptiveMetropolis:
            self.MCMC.use_step_method(pymc.AdaptiveMetropolis, list(
                pymcPars.values()), shrink_if_necessary=True)
        else:
            for par in six.iterkeys(pymcPars):
                self.MCMC.use_step_method(
                    pymc.Metropolis, pymcPars[par], proposal_sd=Steps[par], proposal_distribution='Normal')

        if not "iter" in sampleArgs:
            sampleArgs["iter"] = 2000
        if not "burn" in sampleArgs:
            sampleArgs["burn"] = 0
        if not "thin" in sampleArgs:
            sampleArgs["thin"] = 1

        if not quiet:
            print("Giving the following arguments to 'isample':")
            print("  ", sampleArgs)

        if not quiet:
            self.MCMC.isample(**sampleArgs)
        else:
            self.MCMC.sample(**sampleArgs)
        self.basicStats = self.MCMC.stats()
        if not quiet:
            self._basicStatMCMCOutput(self.basicStats)

        # Setting values to ``best fit values'' (lowest deviance)
        mindex = np.argmin(self.MCMC.trace("deviance")[:])
        if not quiet:
            print("Searching for lowest-deviance solution.")
            print("  Index of lowest-deviance solution: ", mindex)
        for par in six.iterkeys(pymcPars):
            # Weird way of accessing element with index "mindex";
            # prevents slicing/indexing problem in pymc-Trace
            self[par] = (self.MCMC.trace(par)[mindex:mindex + 1])[0]
            if not quiet:
                print("  Parameter: ", par, ", value: ", self[par])
        self.updateModel()
        self.MCMC.db.close()

    def fitEMCEE(self, x=None, y=None, yerr=None, nwalker=None, priors=None, pots=None, scales=None,
                 sampleArgs=None, dbfile="chain.emcee", ps=None, emcp=None, toMD=True):
        r"""
        MCMC sampling using emcee package.

        Sample from the posterior probability distribution using the emcee
        package. By default the likelihood is calculated as -0.5 times the
        model chi-square value.

        The emcee sampler can be accessed via the `emceeSampler` attribute,
        which may be used to continue or manipulate sampling.

        Parameters
        ----------
        nwalker : int, optional
            The number of walker to be used. By default, two times the
            number of free parameters is used.
        scales : dictionary, optional
            The scales argument can be used to control the initial distribution
            of the walkers. By default, all walkers are distributed around the
            location given by the current state of the object, i.e., the current
            parameter values. In each direction, the walker are randomly distributed
            with a Gaussian distribution, whose default standard deviation is one.
            The scales argument can be used to control the width of Gaussians used
            to distribute the walkers.
        sampleArgs : dictionary, optional
            Number controlling the sampling process. Use 'burn' (int) to specify
            the number of burn-in iterations (default is 0). Via 'iters' (int)
            the numbers of iterations after the burn-in can be specified (default 1000).
            The 'process' (int) key can be used to control the number of iterations after
            which the progress bar is updated (default is iters/100). Note that the
            'progressbar' package must be installed to get a progress bar. Otherwise
            more mundane print statements will be used.  
        priors : dictionary, optional
            For each parameter, a primary can be specified. In particular, a
            prior is a callable, which is called with two arguments: first, a
            dictionary mapping the names of the free parameters to their
            current values, and second, a string specifying the name of the
            parameter for which the prior is to apply. The return value must be
            the logarithmic prior probability (natural logarithm). A number of default
            priors are available in the form of the `FuFPrior` class. By
            default, a uniform (improper) prior is used for all parameter, for
            which no other prior was specified.
        pots : list, optional
            A list of 'potentials'. A potential is a function, which is called using
            a dictionary holding the current value for all parameters and returns
            the logarithm of the associated probability. Potentials may, e.g., be
            used to implement certain relations between parameter values not otherwise
            accounted for. 
        dbfile : string, optional
            The result of the sampling, i.e., the chain(s), the corresponding
            values of the posterior, and the names of the free parameters are
            saved to the specified file (by default 'chain.emcee' is used).
            The traces stored there can be analyzed using the 'TraceAnalysis'
            class. Set this parameter to 'None' to avoid saving the results.
        ps : tuple, optional
            A tuple holding the current position and state of the sampler. This
            tuple is returned by this method. The `ps` argument can be used
            to continue sampling from the last state. Note that no burn-in will
            ne carried out and the other arguments should be given as previously
            to continue sampling successfully. 
        emcp : dictionary, optional
            Extra arguments handed to `EnsembleSampler` object.
        toMD : boolean, optional
            If True (default), the object is set to the lowest-deviance solution
            after sampling. Otherwise, it remains in a random state.
        """

        if not ic.check["emcee"]:
            raise (PE.PyARequiredImport("Could not import the 'emcee' package.",
                                        solution="Please install 'emcee'."))

        if (not x is None) and (not y is None) and (not yerr is None):
            # Assign attributes and check x, y, and yerr.
            self._fufDS = FufDS(x, y, yerr)
        elif (not x is None) and (not y is None) and (yerr is None):
            raise (PE.PyAValError("An error on the y values is required.",
                                  where="fitEMCEE",
                                  solution="Please specify 'yerr'"))
        if self._fufDS is None:
            raise (PE.PyAValError("Please specify the data completely.",
                                  where="fitEMCEE",
                                  solution="Specify x, y, and yerr."))

        if not self._fufDS.xyyerrDefined():
            raise (PE.PyAValError("Please specify the data completely. Either of x, y, and/or yerr are missing.",
                                  where="fitEMCEE",
                                  solution="Specify x, y, and yerr."))

        # Names and values of free parameters
        fps = self.freeParameters()
        # Names of the free parameters in specific order
        fpns = self.freeParamNames()
        # Number of dimensions
        ndims = len(fps)

        if ndims == 0:
            raise (PE.PyAValError("At least one free parameter is required for sampling.",
                                  where="fitEMCEE",
                                  solution="Use 'thaw' to free same parameters."))

        if not dbfile is None:
            if re.match(r".*\.emcee$", dbfile) is None:
                PE.warn(PE.PyAValError("The db filename (" + str(dbfile) + ") does not end in .emcee. TraceAnalysis will not recognize it as an emcee trace file.",
                                       solution="Use a filename of the form *.emcee"))

        # Number of walkers
        if nwalker is None:
            self.nwalker = ndims * 2
        else:
            self.nwalker = nwalker

        if self.nwalker < ndims * 2:
            raise (PE.PyAValError("The number of walkers must be at least twice the number of free parameters.",
                                  where="fitEMCEE",
                                  solution="Increase the number of walkers."))
        if self.nwalker % 2 == 1:
            raise (PE.PyAValError("The number of walkers must be even.",
                                  where="fitEMCEE",
                                  solution="Use an even number of walkers."))

        # Use default prior for those parameters not listed
        if priors is None:
            priors = {}
        for n in fpns:
            if not n in priors:
                priors[n] = FuFPrior("uniform")

        # Ensure that potentials is at least an empty list
        if pots is None:
            pots = []

        # Chi square calculator
        chisqr = self.__chiSqr()

        def likeli(names, vals):
            # The likelihood function
            likeli = -0.5 * chisqr(vals)
            return likeli

        def lnpostdf(values):
            # Parameter-Value dictionary
            ps = dict(zip(fpns, values))
            # Check prior information
            prior_sum = 0
            for name in fpns:
                prior_sum += priors[name](ps, name)
            # If log prior is negative infinity, parameters
            # are out of range, so no need to evaluate the
            # likelihood function at this step:
            pdf = prior_sum
            if pdf == -np.inf:
                return pdf
            # Likelihood
            pdf += likeli(fpns, values)
            # Add information from potentials
            for p in pots:
                pdf += p(ps)
            if np.isnan(pdf):
                raise (PE.PyAValError("Posterior value is NaN for parameters: " + str(self.parameters()) + ".",
                                      where="fitEmcee",
                                      solution="Possibly, a prior (e.g., 'limuniform') can be used to restrict parameter range. " +
                                      "Note that restrictions are not automatically converted into priors."))
            return pdf

        # Set default values for sampleArgs
        if sampleArgs is None:
            sampleArgs = {}
        if not "burn" in sampleArgs:
            sampleArgs["burn"] = 0
        if not "iters" in sampleArgs:
            sampleArgs["iters"] = 1000
        if not "progress" in sampleArgs:
            sampleArgs["progress"] = sampleArgs["iters"] / 100

        if ps is None:

            if emcp is None:
                emcp = {}

            # Generate the sampler
            self.emceeSampler = emcee.EnsembleSampler(
                self.nwalker, ndims, lnpostdf, **emcp)

            if scales is None:
                scales = {}

            # Generate starting values
            pos = []
            rrs = self.pars.getRestrictions()
            for _ in smo.range(self.nwalker):
                pos.append(np.zeros(ndims))
                for i, n in enumerate(fpns):
                    if not n in scales:
                        s = 1.0
                    else:
                        s = scales[n]

                    # Trial counter -- avoid values beyond restrictions
                    tc = 0
                    while True:
                        if tc == 100:
                            raise (PE.PyAAlgorithmFailure("Could not determine valid starting point for parameter: " + str(fps) + " due to restrictions",
                                                          where="fitEmcee",
                                                          solution=["Try to use 'scale' to limit range of trial starting values.",
                                                                    "Change starting value before MCMC call into valid range."]))
                        propval = np.random.normal(fps[n], s)
                        if n in rrs:
                            # There is a restriction
                            if (not rrs[n][0] is None) and (propval < rrs[n][0]):
                                tc += 1
                                continue
                            if (not rrs[n][1] is None) and (propval > rrs[n][1]):
                                tc += 1
                                continue
                        break
                    pos[-1][i] = propval

            # Default value for state
            state = None

            if sampleArgs["burn"] > 0:
                # Run burn-in
                pos, prob, state = self.emceeSampler.run_mcmc(
                    pos, sampleArgs["burn"])
                # Reset the chain to remove the burn-in samples.
                self.emceeSampler.reset()

        else:
            # Assign position and state from previous run
            pos, state = ps

        if (not sampleArgs["progress"] is None) and ic.check["progressbar"]:
            widgets = ['EMCEE progress: ', progressbar.Percentage(), ' ', progressbar.Bar(marker=progressbar.RotatingMarker()),
                       ' ', progressbar.ETA()]
            pbar = progressbar.ProgressBar(
                widgets=widgets, maxval=sampleArgs["iters"]).start()

        n = 0

        # Manage emcee 2/3 incompatibility
        try:
            sit = self.emceeSampler.sample(
                pos, rstate0=state, iterations=sampleArgs["iters"], thin=1, storechain=True)
        except TypeError:
            # emcee 3
            sit = self.emceeSampler.sample(
                pos, rstate0=state, iterations=sampleArgs["iters"], thin=1, store=True)

        for pos, prob, state in sit:
            n += 1
            if (not sampleArgs["progress"] is None) and (n % sampleArgs["progress"] == 0):
                if ic.check["progressbar"]:
                    pbar.update(n)
                else:
                    print("EMCEE: Reached iteration ",
                          n, " of ", sampleArgs["iters"])

        # Save the chain to a file
        if not dbfile is None:
            np.savez_compressed(open(dbfile, 'wb'), chain=self.emceeSampler.chain, lnp=self.emceeSampler.lnprobability,
                                pnames=np.array(fpns, dtype=np.str_))

        if toMD:
            # Set to lowest-deviance (highest likelihood) solution
            indimin = np.argmax(self.emceeSampler.lnprobability)
            for i, p in enumerate(self.freeParamNames()):
                self[p] = self.emceeSampler.flatchain[indimin, i]

        return pos, state

    def _resolveMinAlgo(self, minAlgo, default=None, mAA=None):
        r"""
        Resolve minimization algorithm (minAlgo).

        Parameters
        ----------
        minAlgo : callable, string, or None
            If None, the default will be used. If it is
            a callable, it will be assumed to be a valid
            implementation of a minimization algorithm. If
            it is a string, the function will try to resolve
            it.
        default : callable, string, or None
            The algorithm used, if minAlgo itself is None.
        mAA : dictionary, optional
            Keyword arguments given to constructor of minAlgo.

        Returns
        -------
        minAlgo : callable
            An instance of the minimization algorithm.
        """
        if minAlgo is None:
            # If not specified use default.
            minAlgo = default

        if minAlgo is None:
            # This should not happen. Default must be specified.
            raise (PE.PyAValError("No minimization algorithm specified.",
                                  solution="Use, e.g., minAlgo='spfmin' or minAlgo='fufnm'"))

        if hasattr(minAlgo, '__call__'):
            # It is a callable. Just assume that it can be used
            return minAlgo

        # Check keyword arguments to be handed to constructor
        # of the fitter
        if mAA is None:
            maa = {}
        else:
            maa = mAA

        if isinstance(minAlgo, six.string_types):
            if minAlgo == "fufnm":
                return FuFNM(self, **maa)
            elif minAlgo == "spfmin":
                # scipy.optimize.fmin
                if not _scoImport:
                    raise (PE.PyARequiredImport("SciPy.optimize could not be imported.",
                                                solution=["Install SciPy (see www.scipy.org/).",
                                                          "Use funcFit's Nelder-Mead simplex implementation (minAlgo='fufnm')"]))
                return ScipyFMIN(**maa)
            elif minAlgo == "spfmp":
                # scipy.optimize.fmin
                if not _scoImport:
                    raise (PE.PyARequiredImport("SciPy.optimize could not be imported.",
                                                solution=["Install SciPy (see www.scipy.org/).",
                                                          "Use funcFit's Nelder-Mead simplex implementation (minAlgo='fufnm')"]))
                return ScipyFMIN_POWELL(**maa)
            else:
                raise (PE.PyAValError("Unknown string identifier for minimization algorithm: " + minAlgo,
                                      solution="Use, e.g., minAlgo='spfmin' or minAlgo='fufnm'"))

        else:
            raise (PE.PyAValError("Could not resolve 'minAlgo'.",
                                  where="funcFit::OneDFit",
                                  solution="Use, e.g., minAlgo='spfmin' or minAlgo='fufnm'"))

    def fit(self, x, y, yerr=None, X0=None, minAlgo=None, mAA=None, miniFunc=None, printTime=False, *fminPars, **fminArgs):
        r"""
        Carries out a fit.

        Uses an internal optimizer to find the best-fit parameters.
        By default, the method uses the scipy.optimize.fmin algorithm
        (Nelder-Mead Simplex) to find the best-fit parameters. After the
        fit, the parameter values are set to the best-fit values.

        Parameters
        ----------
        x,y : arrays
            Specify the abscissa and ordinate values of the data points.
        yerr : array, optional
            Error of data values.
        X0 : list, optional
            The initial guess. If not provided, it will be assumed
            that the current parameter values already contains
            the initial guess.
        minAlgo : callable or string, optional
            The minimization algorithm. If not specified, scipy's 'fmin'
            implementation will be used. If a callable is given, it
            must adhere to funcFit's internal optimizer model (a suitable
            example is scipy's fmin_powell).
            Valid strings are:
              - 'spfmin' : scipy.optimize.fmin
              - 'fufnm' : funcFit's implementation of the Nelder-Mead simplex algorithm
              - 'spfmp' : scipy.optimize.fmin_powell
        mAA : dictionary, optional
            Keyword arguments handed to the constructor of minAlgo, i.e.,
            minAlgo Arguments (mAA). Valid keywords depend on the choice
            of the fit algorithm.
        fminArgs : dict 
            Keywords passed to the minimization method
            (e.g., `xtol` or `ftol` for scipy.optimize.fmin).  
        fminPars :
            Non-keyword arguments passed to the  minimization method
            (e.g., fprime in scipy.optimize.fmin_ncg).
        miniFunc : None, string {"chisqr", "cash79"}, or function
            Function to be minimized. If None or "chisqr" is given,
            chi-square will be minimized. If "cash79" is given, the
            Cash statistics (Cash 1979, ApJ 228, 939, Eq. 5) will
            be minimized. If a function is specified, that, potentially
            user defined, function will be used to calculated the
            statistics, which will be minimized.
        printTime: boolean, optional
            If True, the number of seconds needed to carry out the fit
            is printed. Default is False. At any rate, the time in seconds
            is stored in the "requiredTime" attribute.
        """
        # Assign attributes and check x, y, and yerr.
        if (x is not None) and (y is not None):
            # This is a not-so-beautiful way to allow
            # calls from within without redefining the
            # data object. This way, the API can remain
            # unchanged.
            self._fufDS = FufDS(x, y, yerr)
        # Choose minimization algorithm
        self.mAA = copy.copy(mAA)
        self.minAlgo = self._resolveMinAlgo(minAlgo, default="spfmin", mAA=mAA)

        # Determine function to be minimized
        if (miniFunc is None) and (yerr is not None):
            miniFunc = "chisqr"
        elif (miniFunc is None) and (yerr is None):
            miniFunc = "sqrdiff"
        self.setObjectiveFunction(miniFunc)
        # Assign initial guess if necessary
        if X0 is not None:
            self.pars.setFreeParams(X0)
        # Save fminPars and fminArgs to internal variables
        self.fminArgs, self.fminPars = fminArgs['fminArgs'] if 'fminArgs' in fminArgs else {
        }, fminPars
        # Carry out fit
        fitStartTime = timestamp()
        # For "historical" reasons, the fitResult must hold a list of the
        # best-fit parameter values as its first item and the final value of
        # the objective function as its second item.
        self.fitResult = self.minAlgo(self.miniFunc, self.pars.getFreeParams(), *self.fminPars,
                                      **self.fminArgs)

        # Set parameters and model to best-fit values
        self.pars.setFreeParams(self.fitResult[0])
        self.updateModel()
        self._stepparEnabled = True

        self.requiredTime = timestamp() - fitStartTime
        if printTime:
            print("The fit took " + str(self.requiredTime) + " seconds.")

    def __extractFunctionValue(self, fr):
        r"""
        Returns the function value (e.g., chi-square).

        Parameters
        ----------
        fr : list
            The fit result returned by the fit method
            used by the `fit` method.

        Returns
        -------
        Function value : float
            For example, chi-square.
        """
        return fr[1]

    def steppar(self, pars, ranges, extractFctVal=None, quiet=False):
        r"""
        Allows to step a parameter through a specified range.

        This function steps the specified parameters through the given
        ranges. During each steps, all free parameters, except for those
        which are stepped, are fitted. The resulting contours allow
        to estimate confidence intervals.

        This command uses the fitting parameters specified on a call
        to the `fit` method. In particular, the same values for `x`,
        `y`, `yerr`, `minAlgo`, `miniFunc`, `fminPars`, and `fminArgs`
        are used.

        .. note:: You need to have carried out a fit before you can
                  use `steppar`.

        Parameters
        ----------
        pars : string or list of strings
            The parameter(s) which are to be stepped.
        ranges : dictionary
            A dictionary mapping parameter name to range specifier.
            The latter is a list containing [lower limit, upper limit,
            no. of steps, 'lin'/'log']. The fourth entry, which
            is optional, is a string specifying whether a constant
            linear step size ('lin') or a constant logarithmic
            step size ('log') shall be used.
        quiet : boolean, optional
            If True, output will be suppressed.
        extractFctVal : callable, optional
            A function specifying how the function value is extracted
            from the fit result. If standard settings are used, the
            default of None is adequate.

        Returns
        -------
        Parameter steps : list
            The return value is a list of lists. Each individual list
            contains the values of the stepped parameters as the first
            entries (same order as the input `pars` list), the
            following entry is the value of the objective function
            (e.g., chi square), and the last entry is a tuple
            containing the indices of the steps of the parameter values.
            This last entry can be useful to convert the result into
            an arrow to plot, e.g., contours. 
        """
        if not self._stepparEnabled:
            raise (PE.PyAOrderError("Before you can use steppar, you must call a function, which enables its use (e.g., `fit`).",
                                    solution="Call the `fit` method first and then try again."))
        if isinstance(pars, six.string_types):
            # Make it a list
            pars = [pars]
        # Check parameter consistency
        for p in pars:
            # Check existence
            tmp = self[p]
            if not p in ranges:
                raise (PE.PyAValError("There is no range for parameter: " + p,
                                      solution="Specify a range; e.g., {'xyz':[0.5,1.9,20,'lin']}"))
        # Function to extract function value from the fit result
        if extractFctVal is None:
            self._extractFctVal = self.__extractFunctionValue
        else:
            if not hasattr(extractFctVal, "__call__"):
                raise (PE.PyAValError("`extractFctVal` needs to be callable!",
                                      solution="Specify a function here or try to use None."))
            self._extractFctVal = extractFctVal
        # Set up ranges
        rs = []
        for par in pars:
            r = ranges[par]
            if len(r) > 4:
                # Use the axis as given
                rs.append(r)
                continue
            if len(r) < 4:
                # By default, use linear spacing
                mode = 'lin'
            else:
                if not isinstance(r[3], six.string_types):
                    raise (PE.PyAValError("If the range has 4 entries, the fourth must be a string specifying the mode.",
                                          solution="Use either 'lin' or 'log' as the fourth entry."))
                mode = r[3]
            if mode == 'lin':
                rs.append(np.linspace(r[0], r[1], r[2]))
            elif mode == 'log':
                # Calculate factor
                s = np.power((r[1] / r[0]), 1.0 / r[2])
                rs.append(r[0] * np.power(s, np.arange(r[2])))
            else:
                raise (PE.PyAValError("Unknown mode: " + str(mode),
                                      solution="Use either 'lin' or 'log'."))
        # Save state of object
        saveObj = self.saveState()
        saveFitResult = self.fitResult
        saveModel = self.model.copy()
        # Freeze parameters, which are affected
        self.freeze(pars)
        # Store result
        result = []
        # Loop over the axes
        nli = pyaC.NestedLoop(list(map(len, rs)))
        for index in nli:
            for i, p in enumerate(pars):
                self[p] = rs[i][index[i]]
            # Fit using previous setting
            # Note that mAA is dispensable, because self.minAlgo will be a callable.
            self.fit(None, None, yerr=None, minAlgo=self.minAlgo, miniFunc=self.miniFunc,
                     *self.fminPars, **self.fminArgs)
            # Build up result
            ppr = []
            for par in pars:
                ppr.append(self[par])
            try:
                ppr.append(self._extractFctVal(self.fitResult))
            except Exception as e:
                PE.warn(PE.PyAValError("The call to the `extractFctVal` function failed. Using full output." +
                                       "\n  Original message: " + str(e)))
                ppr.append(self.fitResult)
            if not quiet:
                print("Result from last iteration:")
                print("  ", ppr)
            ppr.append(index)
            result.append(ppr)
        # Restore old state of object
        self.restoreState(saveObj)
        self.fitResult = saveFitResult
        self.model = saveModel
        return result

    def errorConfInterval(self, par, dstat=2.706, statTol=1e-2, hardLimit=None, maxiter=100, scale=None):
        r"""
        Calculate confidence interval for a parameter.

        This function uses linear extrapolation (similar
        to Newton's method) to find the points, where the
        objective function reaches an offset of `dstat`
        from the minimum value. 

        .. note:: You need to have carried out a fit before you can
                  use `errorConfInterval`.

        Parameters
        ----------
        par : string
            The parameter name for which to determine the error.
        dstat : float, optional
            The offset from the minimum to be reached in the objective
            function's value. For chi-square, a value of 1.0 corresponds
            to a one sigma (68%) confidence interval whereas the default
            of 2.706 refers to a 90% confidence interval.
        statTol : float, optional
            The acceptable (absolute) tolerance in achieving the
            target value in the objective function. The default is 0.01.
        hardLimit : list of two floats, optional
            Hard limits for the confidence interval. You can specify hard
            limits independently of parameter restrictions defined otherwise.
            To specify limits, specify a list of two floats defining
            the lower and upper limit. Use None where no limit is needed:
            e.g., use hardLimits=[0.0, None].
        maxiter : int, optional
            The maximum number of acceptable iterations. If exceeded,
            the calculation of the confidence interval is stopped and
            None is returned. The default is 100.
        scale : float, optional
            This number should reflect the typical scale or extent of
            the problem. If nothing is specified, a tenth of the current
            parameter value is used.

        Returns
        -------
        Confidence interval : dictionary
            The dictionary contains the following keys:
              - "limits" : A list of two floats specifying the lower and upper
                           limits of the confidence interval.
              - "OFVals" : A list of two floats containing the values of the
                           objective function at the lower and upper limit.
              - "OFMin" : The minimum value of the objective function.
              - "parMinVal" : The parameter value at the minimum value of the
                              objective function (best-fit value).
              - "iters" : The number of iterations needed to reach the result.
                          Note that the function returns None if the maximum
                          number of iterations is exceeded.
        """
        if not self._stepparEnabled:
            raise (PE.PyAOrderError("No data available. Did you carry out a fit already?",
                                    solutions="Fit the data first using `fit` method."))
        # Set up hard limit
        if hardLimit is None:
            hardLimit = [None, None]
        # Check whether hard limits are already violated
        if hardLimit[0] is not None:
            if self[par] < hardLimit[0]:
                raise (PE.PyAValError("The parameter value (" + str(self[par]) + ") is below the lower hard limit" +
                                      " of " + str(hardLimit[0]),
                                      solution="Check parameter values and hard limits."))
        if hardLimit[1] is not None:
            if self[par] > hardLimit[1]:
                raise (PE.PyAValError("The parameter value (" + str(self[par]) + ") is above the upper hard limit" +
                                      " of " + str(hardLimit[1]),
                                      solution="Check parameter values and hard limits."))
        # Set up scale parameter
        if scale is None:
            scale = abs(self[par] * 0.1)
        if scale == 0.0:
            raise (PE.PyAValError("The `scale` is 0.0.",
                                  solution="Provide a scale via the corresponding keyword."))
        # Save state of object
        saveObj = self.saveState()
        saveFitResult = self.fitResult
        saveModel = self.model.copy()
        # Get the minimum value of objective function
        # from fit.
        cvalmin = self.__extractFunctionValue(self.fitResult)
        parmin = self[par]
        iters = 2
        self.freeze(par)
        # Find lower limit
        # Calculate first step
        x0 = parmin
        y0 = cvalmin
        self[par] = parmin - scale
        # Note that mAA is dispensable, because self.minAlgo will be a callable.
        self.fit(None, None, yerr=None, minAlgo=self.minAlgo,
                 miniFunc=self.miniFunc)
        x1 = self[par]
        y1 = self.__extractFunctionValue(self.fitResult)
        while True:
            if y1 < cvalmin:
                PE.warn(PE.PyAParameterConflict(
                    "Found a new minimum for the objective function during error search."))
            iters += 1
            if iters > maxiter:
                PE.warn(PE.PyAValError("Maximum number if iterations (" + str(maxiter) + ") exceeded. Returning None.",
                                       solution=["Increase maximum number of iterations via `maxiter` flag.",
                                                 "Adapt scale parameter via `scale` keyword.",
                                                 "Use `steppar` to explore behavior of objective function."]))
                return None
            # Calculate parameters of straight line connecting points
            m = (y1 - y0) / (x1 - x0)
            b = y1 - m * x1
            # Extrapolate new guess from straight line
            newguess = ((dstat + cvalmin) - b) / m
            # Enforce hard limits
            atLimit = False
            if hardLimit[0] is not None:
                if newguess < hardLimit[0]:
                    newguess = hardLimit[0]
                    atLimit = True
            if hardLimit[1] is not None:
                if newguess > hardLimit[1]:
                    newguess = hardLimit[1]
                    atLimit = True
            self[par] = newguess
            self.fit(None, None, yerr=None, minAlgo=self.minAlgo, miniFunc=self.miniFunc,
                     *self.fminPars, **self.fminArgs)
            # Update guess
            x0 = x1
            y0 = y1
            x1 = self[par]
            y1 = self.__extractFunctionValue(self.fitResult)
            if abs(y1 - (dstat + cvalmin)) < statTol:
                break
            if atLimit and (abs(y1 - cvalmin) < dstat):
                break
        # Save lower limit and value of objective function
        lowerLimit = self[par]
        lowerOF = y1

        # Find upper limit
        x0 = parmin
        y0 = cvalmin
        self[par] = parmin + scale
        self.fit(None, None, yerr=None, minAlgo=self.minAlgo, miniFunc=self.miniFunc,
                 *self.fminPars, **self.fminArgs)
        x1 = self[par]
        y1 = self.__extractFunctionValue(self.fitResult)
        while True:
            iters += 1
            if iters > maxiter:
                PE.warn(PE.PyAValError("Maximum number if iterations (" + str(maxiter) + ") exceeded. Returning None.",
                                       solution=["Increase maximum number of iterations via `maxiter` flag.",
                                                 "Adapt scale parameter via `scale` keyword.",
                                                 "Use `steppar` to explore behavior of objective function."]))
                return None
            m = (y1 - y0) / (x1 - x0)
            b = y1 - m * x1
            newguess = ((dstat + cvalmin) - b) / m
            # Enforce hard limits
            atLimit = False
            if hardLimit[0] is not None:
                if newguess < hardLimit[0]:
                    newguess = hardLimit[0]
                    atLimit = True
            if hardLimit[1] is not None:
                if newguess > hardLimit[1]:
                    newguess = hardLimit[1]
                    atLimit = True
            self[par] = newguess
            self.fit(None, None, yerr=None, minAlgo=self.minAlgo, miniFunc=self.miniFunc,
                     *self.fminPars, **self.fminArgs)
            x0 = x1
            y0 = y1
            x1 = self[par]
            y1 = self.__extractFunctionValue(self.fitResult)
            if abs(y1 - (dstat + cvalmin)) < statTol:
                break
            # Check hard limit
            if atLimit and (abs(y1 - cvalmin) < dstat):
                break
        # Save upper limit and value of objective function
        upperLimit = self[par]
        upperOF = y1
        # Restore old state of object
        self.restoreState(saveObj)
        self.fitResult = saveFitResult
        self.model = saveModel
        return {"limits": [lowerLimit, upperLimit], "OFVals": [lowerOF, upperOF], "OFMin": cvalmin,
                "parMinVal": parmin, "iters": iters}


#  def annealFit(self, x, y, scheduler, yerr=None, miniFunc=None):
#    """
#      Carry out a fit using scipy's anneal function.
#    """
#    raise(PE.PyANotImplemented("This is under construction..."))
#    # Assign internal data properties.
#    self.x = x.copy()
#    self.y = y.copy()
#    if yerr is not None:
#      self.yerr = yerr.copy()
#    else:
#      self.yerr = None
#    # Determine function to be minimized
#    if miniFunc is None:
#      self.miniFunc = self.__chiSqr()
#    else:
#      self.miniFunc = miniFunc
#
#    pyaan = PyAAnneal(self)
#    pyaan.anneal(self.miniFunc, scheduler)
#
#    self.assignValue(pyaan.states["best"]["pars"])
#    self.evaluate(x)
#
#    print pyaan.whichStop

#    print "Results of annealing: "
#    print "  ", self.fitResult
#    # Set parameters to best fit
#    self.pars.setFreeParams(self.fitResult[0])
#    # Calculate the best-fit model
#    self.updateModel()


def sampleEMCEE(fpns, fv0, lnp, largs=None, nwalker=None, scales=None, sampleArgs=None, dbfile="chain.emcee", ps=None, emcp=None):
    r"""
    MCMC sampling from specific density using the emcee package.

    This function may be used to use emcee to sample from any user-specified
    density, which does not have to be normalized. The resulting Markov Chains
    can be analyzed using the trace analysis package.

    Parameters
    ----------
    fpns : list of strings
        Names of parameters for which Markov Chains are constructed. 
    fv0 : dictionary
        A dictionary mapping parameter name to starting value. This
        dictionary may contain any number of additional key-value pairs
    lnp : callable
        A function (or equivalent callable) which returns the (natural)
        logarithm of the (generally unnormalized) density. The
        first (and only mandatory) argument to the function is a dictionary
        holding the current parameter values for which the posterior
        density is to be evaluated. The function may take any number of
        additional keyword arguments, which can be specifies by the `largs`
        parameter.
    largs : dictionary, optional
        A set of additional arguments passed to the `lnp` callable.
    nwalker : int, optional
        The number of walker to be used. By default, four times the
        number of free parameters is used.
    scales : dictionary, optional
        The scales argument can be used to control the initial distribution
        of the walkers. By default, all walkers are distributed around the
        location given by the current state of the object, i.e., the current
        parameter values. In each direction, the walkers are randomly distributed
        with a Gaussian distribution, whose default standard deviation is one.
        The scales argument can be used to control the width of the Gaussians used
        to distribute the walkers.
    sampleArgs : dictionary, optional
        Number controlling the sampling process. Use 'burn' (int) to specify
        the number of burn-in iterations (default is 0). Via 'iters' (int)
        the numbers of iterations after the burn-in can be specified (default 1000).
        The 'process' (int) key can be used to control the number of iterations after
        which the progress bar is updated (default is iters/100). Note that the
        'progressbar' package must be installed to get a progress bar. Otherwise
        more mundane print statements will be used.  
    dbfile : string, optional
        The result of the sampling, i.e., the chain(s), the corresponding
        values of the posterior, and the names of the free parameters are
        saved to the specified file (by default 'chain.emcee' is used).
        The traces stored there can be analyzed using the 'TraceAnalysis'
        class. Set this parameter to 'None' to avoid saving the results.
    ps : tuple, optional
        A tuple holding the current position and state of the sampler. This
        tuple is returned by this method. The `ps` argument can be used
        to continue sampling from the last state. Note that no burn-in will
        ne carried out and the other arguments should be given as previously
        to continue sampling successfully. 
    emcp : dictionary, optional
        Extra arguments handed to `EnsembleSampler` object.

    Returns
    -------
    pos, state : state of emcee sample
        These information may be used to continue the sampling from previous position.
    """

    if not ic.check["emcee"]:
        raise (PE.PyARequiredImport("Could not import the 'emcee' package.",
                                    solution="Please install 'emcee'."))

    # Number of dimensions
    ndims = len(fpns)

    if ndims == 0:
        raise (PE.PyAValError("At least one free parameter is required for sampling.",
                              where="sampleEMCEE",
                              solution="TBD."))

    if not dbfile is None:
        if re.match(r".*\.emcee$", dbfile) is None:
            PE.warn(PE.PyAValError("The db filename (" + str(dbfile) + ") does not end in .emcee. TraceAnalysis will not recognize it as an emcee trace file.",
                                   solution="Use a filename of the form *.emcee"))

    # Number of walkers
    if nwalker is None:
        nwalker = ndims * 4

    if nwalker < ndims * 2:
        raise (PE.PyAValError("The number of walkers must be at least twice the number of free parameters.",
                              where="sampleEMCEE",
                              solution="Increase the number of walkers."))
    if nwalker % 2 == 1:
        raise (PE.PyAValError("The number of walkers must be even.",
                              where="sampleEMCEE",
                              solution="Use an even number of walkers."))

    # Set default values for sampleArgs
    if sampleArgs is None:
        sampleArgs = {}
    if not "burn" in sampleArgs:
        sampleArgs["burn"] = 0
    if not "iters" in sampleArgs:
        sampleArgs["iters"] = 1000
    if not "progress" in sampleArgs:
        sampleArgs["progress"] = sampleArgs["iters"] / 100

    # Dictionary used to call the provided log posterior function
    kvs = copy.copy(fv0)

    # Manage None is argument to largs
    if largs is None:
        _largs = {}
    else:
        _largs = largs

    # Generate log posterior function required by emcee
    def logpost(x):
        r"""
        Calls user-specified log posterior function with
        properly updated parameter dictionary.
        """
        for i, n in enumerate(fpns):
            # Assign parameter values, which are variable
            kvs[n] = x[i]
        return lnp(kvs, **_largs)

    if emcp is None:
        emcp = {}

    if ps is None:

        # Generate the sampler
        emceeSampler = emcee.EnsembleSampler(nwalker, ndims, logpost, **emcp)

        if scales is None:
            scales = {}

        # Generate starting values
        pos = []
        for _ in smo.range(nwalker):
            pos.append(np.zeros(ndims))
            for i, n in enumerate(fpns):
                if not n in scales:
                    s = 1.0
                else:
                    s = scales[n]
                pos[-1][i] = np.random.normal(fv0[n], s)

        # Default value for state
        state = None

        if sampleArgs["burn"] > 0:
            # Run burn-in
            pos, prob, state = emceeSampler.run_mcmc(pos, sampleArgs["burn"])
            # Reset the chain to remove the burn-in samples.
            emceeSampler.reset()

    else:
        # Generate the sampler
        emceeSampler = emcee.EnsembleSampler(nwalker, ndims, logpost, **emcp)
        # Assign position and state from previous run
        pos, state = ps

    if (not sampleArgs["progress"] is None) and ic.check["progressbar"]:
        widgets = ['EMCEE progress: ', progressbar.Percentage(), ' ', progressbar.Bar(marker=progressbar.RotatingMarker()),
                   ' ', progressbar.ETA()]
        pbar = progressbar.ProgressBar(
            widgets=widgets, maxval=sampleArgs["iters"]).start()

    n = 0
    for pos, prob, state in emceeSampler.sample(pos, rstate0=state, iterations=sampleArgs["iters"], thin=1):
        n += 1
        if (not sampleArgs["progress"] is None) and (n % sampleArgs["progress"] == 0):
            if ic.check["progressbar"]:
                pbar.update(n)
            else:
                print("EMCEE: Reached iteration ",
                      n, " of ", sampleArgs["iters"])

    # Save the chain to a file
    if not dbfile is None:
        np.savez_compressed(open(dbfile, 'wb'), chain=emceeSampler.chain, lnp=emceeSampler.lnprobability,
                            pnames=np.array(fpns, dtype=np.str_))

    return pos, state



from __future__ import print_function, division
from PyAstronomy.pyaC import ImportCheck as _ImportCheck
from PyAstronomy.pyaC import pyaErrors as _PE
import PyAstronomy as _PyA
from PyAstronomy.constants import _systems
import os
from six.moves import configparser as _ConfigParser
import numpy
import six

_ic = _ImportCheck(["quantities"], ["quantities"])

import quantities as _pq


def _displaySummary(scope):
  """
    Display unit system and constants.
    
    Parameters
    ----------
    scope : module or class
        The scope from which to get the data to
        be displayed.
  """
  print("Current unit system: ", getattr(scope, "getSystem")())
  print()
  
  cmod = scope
  inventory = getattr(scope, "inventory")
    
  maxlenSym = 0
  maxlenDescr = 0
  maxlenVal = 0
  for k in six.iterkeys(inventory):
    maxlenSym = max(maxlenSym, len(k))
    maxlenDescr = max(maxlenDescr, len(inventory[k]["descr"]))
    maxlenVal = max(maxlenVal, len("%10.5e %s" % (getattr(cmod, "f_"+k).magnitude, getattr(cmod, "f_"+k).dimensionality)))
  maxlenSym = max(maxlenSym, len("Symbol"))
  maxlenDescr = max(maxlenDescr, len("Description"))
  maxlenVal = max(maxlenVal, len("Value"))
  
  totlen = maxlenSym + maxlenDescr + maxlenVal + 6
  
  print(("%" + str(maxlenSym) + "s | %" + str(maxlenDescr) + "s | %" + \
         str(maxlenVal) + "s") % ("Symbol", "Description", "Value"))
  print("-" * maxlenSym + "---" + '-' * maxlenDescr + "---" + '-' * maxlenVal)

  for k in sorted(inventory.keys()):
    print((("%" + str(maxlenSym) + "s | ") % k) + \
          ("%" + str(maxlenDescr) + "s") % inventory[k]["descr"] + " | " + \
          "%10.5e %s" % (getattr(cmod, "f_"+k).magnitude, getattr(cmod, "f_"+k).dimensionality))
  print("-" * totlen)


class PyAConstants:
  """
    Class scope for constants.
    
    Parameters
    ----------
    fn : string, optional
        If given, constants will be read from the specified file.
    unitSystem : string, optional, {"SI", "cgs"}
        The system of units to be used. Default is "cgs".
  """
  
  def _inventoryToScope(self):
    """
      Translate information in `inventory` into class attributes.
    """
    for k, u in six.iteritems(self.inventory):
      val = _pq.Quantity(numpy.float64(u["valueSI"]), u["units"]["SI"])
      value = val.rescale(u["units"][self._unitSystem])
      setattr(self, u["symbol"], value.magnitude)
      setattr(self, "f_" + u["symbol"], value)
      err = _pq.Quantity(numpy.float64(u["errSI"]), u["units"]["SI"])
      err = err.rescale(u["units"][self._unitSystem])
      setattr(self, "f_" + u["symbol"] + "_err", err)

  def summary(self):
    """
      Print a summary of available constants to screen.
    """
    _displaySummary(self)

  def setSystem(self, system):
    """
      Set the unit system.
      
      Parameters
      ----------
      system : string, {SI, cgs}
          The unit system in use.
    """
    if not system in _systems:
      raise(_PE.PyAValError("No such system: '" + system + "'.", \
                            solution="Choose one of: "+str(_systems)))
    self._unitSystem = system
    self._inventoryToScope()

  def inUnitsOf(self, const, units):
    """
      Get a constant and apply unit conversion.
      
      Parameters
      ----------
      const : string
          Name of constant.
      units : string
          The units to be converted into.
      
      Returns
      -------
      Constant : Quantity object (`quantities` package)
          The value and unit of the constant.
    """
    if not hasattr(self, "f_" + const):
      raise(_PE.PyAValError("No such constant: '" + str(const) + "'"))
    val = getattr(self, "f_" + const)
    return val.rescale(units)

  def cleanUp(self):
    """
      Clean up loaded constants.
      
      Empties the `inventory` and deletes all
      related module attributes.
    """
    for k, u in six.iteritems(self.inventory):
      delattr(self, u["symbol"])
      delattr(self, "f_" + u["symbol"])
      delattr(self, "f_" + u["symbol"] + "_err")
    self.inventory = {}

  def constantDetails(self, const):
    """
      Print all information for a specific constant.
      
      Parameters
      ----------
      const : string
          The constant.
      
      Returns
      -------
      info : dictionary
          A dictionary containing all information about the
          specified constant.
    """
    if not const in self.inventory:
      raise(_PE.PyAValError("No such constant: '" + str(const) + "'."))
    print("Constant              : '" + const + "'")
    print("Description           : " + self.inventory[const]["descr"])
    print("Current value and unit: " + str(getattr(self, "f_"+const)))
    print("Uncertainty           : " + str(getattr(self, "f_"+const+"_err")))
    print("Source                : " + self.inventory[const]["source"])
    return self.inventory[const]

  def getSystem(self):
    """
      Get current unit system.
      
      Returns
      -------
      System : string, {SI, erg}
          The unit system currently in use.
    """
    return self._unitSystem

  def load(self, fn, constClobber=False):
    """
      Load constants file.
      
      Parameters
      ----------
      fn : string
          The file name.
      constClobber : boolean, optional
          If True, existing constants will be overwritten.
    """
    config = _ConfigParser.RawConfigParser()
    config.read(fn)
    for s in config.sections():
      nc = None
      try:
        nc = {}
        nc["symbol"]  = config.get(s, "symbol")
        nc["descr"]   = config.get(s, "descr")
        nc["valueSI"] = config.get(s, "valueSI")
        nc["errSI"]   = config.get(s, "errSI")
        nc["units"] = {}
        for u in _systems:
          nc["units"][u] = config.get(s, "unit"+u)
        nc["source"] = config.get(s, "source")
      except _ConfigParser.NoOptionError as e:
        _PE.warn(_PE.PyAValError("The definition of the constant defined in section '" + s + "' of file '" + \
                                 fn + "' is incomplete. Ignoring this constant. Caught the error:\n" + \
                                 str(e)))
        nc = None
      if nc is not None:
        if (nc["symbol"] in self.inventory) and (not constClobber):
          raise(_PE.PyANameClash("A constant with the symbol '" + nc["symbol"] + "' is already defined.", \
                                 solution=["Choose another symbol.", "Set the `constClobber` flag to True to overwrite existing definitions of constants."])) 
        self.inventory[nc["symbol"]] = nc
      self._inventoryToScope()
    
  
  def __init__(self, fn=None, unitSystem="cgs"):
    if fn is None:
      # Use default file.
      fn = os.path.join(os.path.dirname(_PyA.__file__), "constants/cdat.dat")
    self.inventory = {}
    self._unitSystem = unitSystem
    self.load(fn)
    



from PyAstronomy.pyaC import pyaErrors as PE
import re


def hmsToDeg(h, m, s):
    r"""
    Convert hour-minute-second specification into degrees.

    Parameters
    ----------
    h : float
        Hours (0-24)
    m : float
        Minutes (time, 0-60)
    s : float
        Seconds (time, 0-60)

    Returns
    -------
    Angle : float
        The corresponding angle in degrees.
    """
    if (h < 0.0) or (h >= 24.0):
        raise (PE.PyAValError("Hour (" + str(h) + ") out of range (0 <= h < 24).",
                              solution="Specify a value between 0 and 24."))
    if (m < 0.0) or (m >= 60.0):
        raise (PE.PyAValError("Minute (" + str(m) + ") out of range (0 <= m < 60)",
                              solution="Specify a value between 0 and 60."))
    if (s < 0.0) or (s >= 60.0):
        raise (PE.PyAValError("Second (" + str(s) + ") out of range (0 <= m < 60)",
                              solution="Specify a value between 0 and 60."))
    result = h * 15.0 + m * 0.25 + s * (0.25 / 60.0)
    return result


def degToHMS(d):
    r"""
    Convert degrees into time units (hours-minutes-seconds)

    Parameters
    ----------
    d : float
        Degrees (0-360)

    Returns
    -------
    h, m, s: float
        Hours, minutes, and seconds
    """
    if (d < 0.0) or (d >= 360.0):
        raise (PE.PyAValError("The argument should be between 0 and <360 degrees."))
    h = int(d / 15.0)
    d = d - h * 15.0
    m = int(d / 0.25)
    d = d - m * 0.25
    s = d / (0.25 / 60.0)
    return h, m, s


def degToDMS(g):
    r"""
    Convert degrees into arc units (degrees, (arc)minutes, (arc)seconds)

    Parameters
    ----------
    g : float
        Value in degrees.

    Returns
    -------
    d, m, s, sign : float, int
        Degrees, (arc)minutes, and (arc)seconds. Note that only the
        degree number is signed. The sign (+1/-1) is also returned to
        yield a complete result if the value of degree (d) is zero. 
    """
    sign = 1
    if g < 0.0:
        sign = -1
    g = abs(g)
    d = int(g) * sign
    g = g - int(g)
    m = int(g * 60.0)
    g = g - m / 60.0
    s = g * 3600.0
    return d, m, s, sign


def dmsToDeg(d, m, s, esign=0):
    r"""
    Convert degree-arcminute-arcsecond specification into degrees.

    Parameters
    ----------
    d : float
        Degrees.
    m : float
        Arcminutes (0-60)
    s : float
        Arcseconds (0-60)
    esign : int, optional, {-1,0,1}
        Explicit sign with -1 representing negative sign, +1 representing
        positive sign, and 0 indicating no explicit sign specification.
        The explicit sign is necessary if negative southern coordinates are
        specified but d is 0 and, thus, cannot carry the sign. 

    Returns
    -------
    Angle : float
        The corresponding angle in degrees.
    """
    if (m < 0.0) or (m >= 60.0):
        raise (PE.PyAValError("Minute (" + str(m) + ") out of range (0 <= m < 60)",
                              solution="Specify a value between 0 and 60."))
    if (s < 0.0) or (s >= 60.0):
        raise (PE.PyAValError("Second (" + str(s) + ") out of range (0 <= m < 60)",
                              solution="Specify a value between 0 and 60."))
    if not esign in (-1, 0, +1):
        raise (PE.PyAValError("Invalid value for 'esign' (value given is " + str(esign) + ").",
                              solution="Use +1, 0, or -1.",
                              where="dmsToDeg"))
    if (d == 0.0) and (esign == 0):
        raise (PE.PyAValError("When a value of zero is specified for degrees (d), the sign has to be given explicitly via the `esign` flag.",
                              where="dmsToDeg",
                              solution="Specify esign=+/-1 according to sign (i.e., northern or southern coordinates).",
                              addInfo="Safety device added to prevent common conversion errors from coordinates like '-00 12 08'."))

    sign = 1.0
    if d < 0.0:
        sign = -1.0

    if (d != 0.0) and (esign != 0):
        # Check consistency of sign and esign
        if int(sign) != esign:
            raise (PE.PyAValError("The sign of 'd' and the specification of 'esign' must not contradict.",
                                  solution="Adjust 'esign' specification.",
                                  where="dmsToDeg"))
    elif d == 0.0:
        sign = esign

    result = d + sign * (m / 60.0) + sign * (s / 3600.0)
    return result


def coordsSexaToDeg(c, fullOut=False):
    r"""
    Convert sexagesimal coordinate string into degrees.

    Parameters
    ----------
    c : string
        The coordinate string. Valid formats are, e.g.,
        "00 05 08.83239 +67 50 24.0135" or "00:05:08.83239 -67:50:24.0135".
        Spaces or colons are allowed as separators for the individual
        components of the coordinates.
    fullOut : boolean, optional
        If True, two additional tuples holding the individual components of
        the right ascension and declination specification will be returned.
        The default is False.

    Returns
    -------
    ra, dec : float
        Right ascension and declination in degrees.
    hms, dms : tuples of three floats
        If `fullOut` is True, two tuples of three numbers each will be
        returned, which hold the individual constituents making up the
        right ascension (hms) and declination (dms) specifiers in sexagesimal
        representation.
    """
    r = re.match(
        r"^\s*(\d+)([\s:])(\d+)([\s:])(\d+(\.\d*)?)\s+(([+-])(\d+))([\s:])(\d+)([\s:])(\d+(\.\d*)?)\s*$", c)
    if r is None:
        raise (PE.PyAValError("Could not decompose coordinate string: \"" + str(c) + "\"",
                              solution="Use, e.g., 00 05 08.83239 +67 50 24.0135 or 00:05:08.83239 +67:50:24.0135"))
    # Check separator consistency
    for n in [4, 10, 12]:
        if r.group(2) != r.group(n):
            raise (PE.PyAValError("Separators (space and colon) mixed in coordinate definition.",
                                  solution="Use consistent separator."))

    ra = (float(r.group(1)), float(r.group(3)), float(r.group(5)))
    esign = {'+': +1, '-': -1}[r.group(8)]
    de = (float(r.group(7)), float(r.group(11)), float(r.group(13)), esign)
    result = (hmsToDeg(*ra), dmsToDeg(*de))
    if not fullOut:
        return result
    return result[0], result[1], ra, de


def coordsDegToSexa(ra, dec, asString=True, fmt=("%02d %02d %06.3f  ", "%s%02d %02d %06.3f")):
    r"""
    Convert right ascension and declination from degrees into sexagesimal representation.

    Parameters
    ----------
    ra : float
        Right ascension in degrees.
    dec : float
        Declination in degrees.
    asString : boolean, optional
        If True (default), the result will be a string formatted
        according to the rules specified by `fmt`.
    fmt : tuple of strings, optional
        The output format used to create the output string (first ra, second dec). Only used
        if `asString` is True (default).

    Returns
    -------
    Coordinates : string or tuple
        If `asString` is True (default), a string holding the coordinates
        is returned, which is formatted according to the rules specified by
        the `fmt` parameter. If False, a tuple of two tuples is returned,
        of which the first holds three numbers representing the right ascension
        (hms) and the second three numbers representing the declination (dms, sign).    
    """
    if (ra < 0.0) or (ra >= 360.):
        raise (PE.PyAValError("Right ascension should be between 0 and 360 degrees."))
    if (dec < -90.0) or (dec > 90.0):
        raise (PE.PyAValError("Declination should be between -90 and +90 degrees."))
    rat = degToHMS(ra)
    dect = degToDMS(dec)

    if asString:
        result = (fmt[0] % rat) + (fmt[1] %
                                   ({-1: '-', +1: '+'}[dect[3]], abs(dect[0]), dect[1],  dect[2]))
        return result
    return rat, dect




import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # Model evaluation
        # =============================
        
        # In[1]:
        
        
        from PyAstronomy import funcFit2 as fuf2
        import numpy as np
        import matplotlib.pylab as plt
        import matplotlib
        matplotlib.use('Agg')
        
        # Create a a model object representing a Gaussian
        gf = fuf2.GaussFit()
        
        # Parameters can be accessed using the square bracket
        # notation
        gf["A"] = 10.0
        gf["sig"] = 15.77
        gf["off"] = 1.0
        gf["mu"] = 7.5
        
        x = np.linspace(gf["mu"]-5*gf["sig"], gf["mu"]+5*gf["sig"], 150)
        y = gf.evaluate(x)
        
        gf["A"] = 7.5
        y2 = gf.evaluate(x)
        
        plt.plot(x, y, 'b.-', label="A=10")
        plt.plot(x, y2, 'r.-', label="A=7.5")
        plt.legend()
        plt.show()
        
        # In[2]:
        
        
        
        



from PyAstronomy.pyaC import _ic
from PyAstronomy.pyaC import pyaErrors as PE
import numpy as np


def ibtrapz(x, y, x0, x1, iaout=False):
    """
      Use the trapezoid rule to integrate and interpolate boundary values.

      Can be used for integration on tabled data, where the integration boundaries
      are not located on the grid. The values to use at the boundaries are
      determined using linear interpolation.

      Parameters
      ----------
      x,y : arrays
          The data.
      x0, x1 : float
          The integration boundaries.
      iaout : boolean, optional
          If True, the arrays used internally for
          integration are also returned. The default
          is False.

      Returns
      -------
      Integral : float
          The value of the resulting integral.
      xi, yi : arrays, optional
          Internally used arrays for integration including
          the values derived at the boundaries. Only returned
          if `iaout` is set to True.
    """
    if not _ic.check["scipy"]:
        raise (PE.PyARequiredImport("scipy could not be imported.",
                                    where="ibtrapz",
                                    solution="Please install scipy to use ibtrapz."))
    import scipy.interpolate as sci
    import scipy.integrate as scinteg
    if (x1 < x0):
        raise (PE.PyAValError("x1 must be larger than x0.",
                              where="ibtrapz",
                              solution="Exchange boundaries."))
    if np.any(x[1:]-x[0:-1] < 0.0):
        raise (PE.PyAValError("x axis must be in ascending order.",
                              where="ibtrapz",
                              solution="Sort values."))
    if (x0 < np.min(x)) or (x1 > np.max(x)):
        raise (PE.PyAValError("One or both integration boundaries are beyond the range of data given.",
                              where="ibtrapz",
                              solution="Give more data or adjust integration boundaries."))
    # Interpolate at integration boundaries
    fi = sci.interp1d(x, y)
    y0, y1 = fi(x0), fi(x1)
    # Search data within boundaries
    indi = np.where((x > x0) & (x < x1))[0]
    # Construct arrays to be used in integration
    xi = np.zeros(len(indi)+2)
    yi = np.zeros(len(indi)+2)
    xi[1:-1] = x[indi]
    yi[1:-1] = y[indi]
    xi[0] = x0
    xi[-1] = x1
    yi[0] = y0
    yi[-1] = y1

    if iaout:
        return scinteg.trapezoid(yi, xi), xi, yi
    return scinteg.trapezoid(yi, xi)




import unittest
class TestSanityOfNotebook(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass
        
    def testsanity_example(self):
        #!/usr/bin/env python
        # coding: utf-8
        
        # # Using the spectral library
        
        # In[1]:
        
        
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import pyasl
        import os
        import matplotlib
        matplotlib.use('Agg')
        
        sl = pyasl.SpectralLib()
        
        # List available libraries
        sl.listDescription()
        
        # Disk-integrated spectrum for Teff, logg, and metallicty
        sdifn = sl.requestModel(5750, 4.5, 0.0, nex="diskint")
        
        print("Model filename: ", os.path.basename(sdifn))
        
        w, f = sl.read1dFitsSpec(sdifn)
        
        plt.xlabel("Wavelength [$\AA$]")
        plt.ylabel("Astrophysical flux [erg/cm**2/s/A]")
        plt.plot(w, f, 'b.-')
        plt.show()
        
        # ## Limb-angle resolved Halpha line
        
        # In[2]:
        
        
        wlmin = 6564-20
        wlmax = 6564+20
        indi = (w > wlmin) & (w < wlmax)
        
        # Specific intensity resolved w.r.t. viewing angle
        # for Teff, logg, and metallicty
        musfn = sl.requestModel(5750, 4.5, 0.0, nex="muspecs")
        
        w, mus, intens = sl.readMuFits(musfn)
        
        print("Available cos(theta) = mu angles: ", mus)
        
        # Plot specific intensity
        for i, mu in enumerate(mus):
            plt.plot(w[indi], intens[indi,i], label="mu = %3.1f" % mu)
        plt.xlabel("Wavelength [$\AA$]")
        plt.ylabel("Specific intensity")
        plt.legend()
        plt.show()
        
        # ## Manuel disk integration
        
        # In[3]:
        
        
        # Real (not astrophysical) flux
        rf = np.zeros_like(w)
        
        for i in range(len(mus)-1):
            # rf += I*mu*dmu
            rf += (intens[::,i]*mus[i] + intens[::,i+1]*mus[i+1])/2 * (mus[i+1] - mus[i])
        rf *= 2*np.pi
        # Astrophysical flux
        af = rf/np.pi
        
        plt.plot(w[::10], f[::10], 'g:', label="Disk-integrated library flux")
        plt.plot(w[::10], af[::10], 'b.:', label="Astrophysical flux (integral)")
        plt.plot(w[::10], pyasl.planck(5750, w[::10]*1e-10)*1e-7/np.pi, 'r--', label="Planck") 
        plt.legend()
        plt.show()
        
        # ## Listing available content
        
        # In[4]:
        
        
        sl.listInventory()
        
        # In[5]:
        
        
        
        



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import numpy
from .onedfit import OneDFit
import six.moves as smo


class PolyFit1d(OneDFit):
    r"""
    Implements a polynomial fit.

    *Fit parameters*:
      - cn - Here n is a number indicating degree
             (e.g., c0 + c1*x + c2*x**2 ...)

    If `xoff` is specified, the polynomial will be
    evaluated at the points `x-xoff`. This can be
    useful to suppress correlation.
    """

    def __init__(self, degree, xoff=0.0):
        coeffs = []
        self.degree = degree
        self.xoff = xoff
        for i in smo.range(degree+1):
            coeffs.append("c"+str(i))
        OneDFit.__init__(self, coeffs)
        self.setRootName("Poly")

    def asNPPoly(self):
        r"""
        Construct a numpy.poly1d object from the coefficients.

        Returns
        -------
        Polynomial : numpy.poly1d
            A numpy polynomial with the coefficients stored
            by the fit object. 
        """
        poly = []
        for i in smo.range(self.degree, -1, -1):
            poly.append(self["c"+str(i)])
        return numpy.poly1d(poly)

    def evaluate(self, x):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        return self.asNPPoly()(x-self.xoff)


class ConstantFit1d(OneDFit):
    r"""
    Implements a constant model.

    *Fit parameters*:
      - `c` - The constant.
    """

    def __init__(self):
        OneDFit.__init__(self, ["c"])
        self.setRootName("Constant")

    def evaluate(self, x):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        return numpy.ones(x.size) * self["c"]



from . import pyaErrors as PE
import numpy as np

def invertIndexSelection(a, indi):
  """
    Invert index selection in one-dimensional array.
    
    Say, e.g., a numpy.where operation produced an array
    of indices (`indi`), which you needed for one reason,
    but now, for another reason, you need all elements of
    the array, which were not selected by the operation.
    This is the situation handled by this function. 
    
    Parameters
    ----------
    a : int or array
        Either the length of the array to which `indi` refers
        to or the array (one-dimensional) itself.
    indi : array
        An array of indices selected from `a`.
    
    Returns
    -------
    Inverse selection : array
        An array containing the indices of all array elements
        not referenced by `indi`. 
  """
  if isinstance(a, int):
    n = a
  elif isinstance(a, np.ndarray):
    if len(a.shape) != 1:
      raise(PE.PyAValError("Found first parameter to be a numpy array, but it is not of dimension one!", \
            solution=["Specify a one-dimensional array", "Specify merely the length of the array as an int."], \
            where="invertIndexSelection"))
    n = len(a)
  else:
    raise(PE.PyAValError("First parameter is neither a one-dimensional numpy array or an integer.", \
          solution="Use one of these types to describe the array."))
  
  x = np.ones(n, dtype=bool)
  x[indi] = False
  return np.arange(n)[x]



from pathlib import Path

fns = list(Path(".").rglob("ex_*.ipynb"))

print("Files matching criteria:")
for i, fn in enumerate(fns,1):
    print(i, fn)
print()


for fn in fns:
    
    ec = 1
    ll = open(str(fn)).readlines()
    for i, l in enumerate(ll):
        if l.find('"execution_count":') != -1:
            ll[i] = l.split(":")[0] + ": %d,\n" % ec
            ec += 1

    open(str(fn), 'wt').writelines(ll)





# -*- coding: utf-8 -*-
from __future__ import print_function, division
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo
import six
from six.moves import configparser as ConfigParser
import os

def observatory(obsName):
  """
    Get the location of an observatory.
    
    This function emulates the behavior of
    the IDL Astrolib's `observatory.pro`
    procedure. Use `showObservatories`
    to get a list of known observatories.
    
    Parameters
    ----------
    obsName : string
        The observatory code. Use `listObservatories`
        to get a list of available observatories
        and their codes.
    
    Returns
    -------
    Observatory location : dictionary
        The keys of this dictionary are: "longitude",
        "latitude", "altitude", and "tz". Longitude
        and latitude are given in degrees and the
        observatory's altitude in meter. Finally, tz
        defines the time zone (hours West of Greenwich).
  """
  config = ConfigParser.RawConfigParser()
  config.read(os.path.join(os.path.dirname(__file__), 'observatory.cfg'))
  
  
  if not config.has_section(obsName):
    raise(PE.PyAValError("No such observatory: '" + str(obsName) + "'", \
          solution="Choose existing observatory (use listObservatories" + \
          " to get a list)", \
          where="observatory"))
  
  result = {}
  result["name"] = config.get(obsName, "NAME")
  result["longitude"] = config.getfloat(obsName, "LONGITUDE")
  result["latitude"] = config.getfloat(obsName, "LATITUDE")
  result["altitude"] = config.getfloat(obsName, "ALTITUDE")
  result["tz"] = config.getfloat(obsName, "TZ")
  
  return result


def listObservatories(show=True):
  """
    Get a list of available observatories.
    
    Parameters
    ----------
    show : boolean, optional
        If True (default), the observatory data
        will be written to screen.
    
    Returns
    -------
    Observatory data : dictionary
        For every observatory code (key), the dictionary
        holds another dictionary with "longitude",
        "latitude", "altitude", "name", and "tz". Longitude
        and latitude are given in degrees and the
        observatory's altitude in meter. Finally, tz
        defines the time zone (hours West of Greenwich)
        and "name" gives the full observatory name.
  """
  config = ConfigParser.RawConfigParser()
  config.read(os.path.join(os.path.dirname(__file__), 'observatory.cfg'))
  codes = config.sections()
  
  result = {}
  for code in codes:
    result[code] = observatory(code)
  
  if show:
    print("List of available observatories")
    print()
    maxCodeLen = max([len(k) for k in six.iterkeys(result)])
    print(("{0:"+str(maxCodeLen)+"s}     ").format("Code") + "Observatory name")
    print("-" * (21+maxCodeLen))
    for k in sorted(list(result), key=lambda s: s.lower()):
      print(("{0:"+str(maxCodeLen)+"s} --- ").format(k) + result[k]["name"])
    print()
    print("Observatory location")
    print("--------------------")
    for k in sorted(list(result), key=lambda s: s.lower()):
      print(("Code: {code:"+str(maxCodeLen)+"s}, " + \
             "Longitude = {longitude:7.3f}°, Latitude = {latitude:+7.3f}°, " + \
            "Altitude = {altitude:5.0f} m, TZ = {tz: 5.1f}").format(code=k, **result[k]))
  return result 


from __future__ import print_function, division
import re
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
import six

class LineOfSight:
  """
    Define the "line of sight".
    
    Constructs a line of sight, i.e., a unit vector in the
    given direction.
    
    Parameters
    ----------
    los : string, {'+x','-x','+y','-y','+z','-z'}, or array
        The line of sight given in string format.
    
    Attributes
    ----------
    
  """
  
  def __init__(self, los='-z'):

    if isinstance(los, six.string_types):
      # The input is a string. Check the content.
      r = re.match("^([-+])([xyz])$", los)
      if r is None:
        raise(PE.PyAValError("A line of sight encoded by " + los + " cannot be constructed."))
      if r.group(2) == "x":
        self.los = np.array([1.0,0.0,0.0])
      elif r.group(2) == "y":
        self.los = np.array([0.0,1.0,0.0])
      elif r.group(2) == "z":
        self.los = np.array([0.0,0.0,1.0])
      
      if r.group(1) == "-":
        self.los *= -1.0
      
      self.losName = los
      return
    
    # Try to convert input to numpy array
    convlos = np.array(los)
    # Check length
    if len(convlos) != 3:
      raise(PE.PyAValError("A line of sight argument does not have the right length of 3."))
    # Check that it is a unit vector
    if np.abs(np.sum(convlos**2) - 1.0) > 1e-8:
      raise(PE.PyAValError("A line of sight argument is not a unit vector."))
    
    self.los = convlos
    self.losName = str(convlos)
      
      
  
  def __str__(self):
    """
      String output of the class.
    """
    return "Line of sight (" + self.losName + "): " + str(self.los)


from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo

class SVD:
  """
    Use Singular Value Decomposition (SVD) to obtain broadening function.
    
    The technique is, e.g., described in Rucinski 1999, "Determination of
    Broadening Functions Using the Singular-Value Decomposition (SVD)
    Technique" and references therein.
  """
  
  def __init__(self):
    self.w = None
    self.des = None
    self.bn = None
  
  def _createDesignMatrix(self, template, bn):
    """
      Creates the "design matrix".
      
      The design matrix is written to the attribute `des`.
      
      Parameters
      ----------
      template : array
          The template spectrum.
      bn : int
          The length (number of bins) used for the
          broadening function.
    """
    # Number of bins used for the design matrix (edges
    # of the template are cut)
    n = len(template) - bn + 1
    self.des = np.matrix(np.zeros( (bn, n) ))
    for i in smo.range(bn):
      self.des[i,::] = template[i:i+n]
    self.des = self.des.T
  
  def getSingularValues(self):
    """
      Access the singular values.
      
      Returns
      -------
      Singular values : matrix
          The singular values.
    """
    if self.w is None:
      raise(PE.PyAOrderError("No singular values available yet.", \
                             "Use the `decompose` method first."))
    return self.w[:]
  
  def getModel(self, broad, asarray=False, modelIndices=False):
    """
      Calculates a model resulting from template and broadening function.
      
      Parameters
      ----------
      broad : array or matrix
          The broadening function
      asarray : bool, optional
          If True, the broadening function will be returned
          as an array instead of a matrix. The default is
          False.
      modelIndices : bool, optional
          If True, the method returns also an array of indices,
          which refer to the template, for which the model is
          calculated. Note that the model cannot be calculated
          at the edges. The default is False.
    """
    if self.des is None:
      raise(PE.PyAOrderError("You need to create a design matrix first.", \
                             solution="Use the `decompose` method."))
    if not modelIndices:
      if not asarray:
        return self.des * np.matrix(broad)
      else:
        return (self.des * np.matrix(broad)).getA()[::,0]
    else:
      modelInd = np.arange(self.bn//2, len(self.template)-self.bn//2)
      if not asarray:
        return (self.des * np.matrix(broad), modelInd)
      else:
        return ((self.des * np.matrix(broad)).getA()[::,0], modelInd)
  
  def getBroadeningFunction(self, flux, wlimit=0.0, asarray=False):
    """
      Calculate the broadening function.
      
      .. note:: The `decompose` method has to be called
                first. On this call, the template is
                specified.
      
      Parameters
      ----------
      flux : array or matrix
          The observed function (flux).
      wlimit : float, optional
          A limit to be applied to the singular
          values. Values smaller than the specified
          limit will be discarded in calculating
          the broadening function.
      asarray : bool, optional
          If True, the broadening function will be returned
          as an array instead of a matrix. The default is
          False.
      
      Returns
      -------
      Broadening function : matrix, array
          The broadening function. The return type can be set
          by the `asarray` flag.
    """
    if len(flux) == (len(self.template) - self.bn + 1):
      validIndi = np.arange(len(self.template) - self.bn + 1)
    elif len(flux) == len(self.template):
      validIndi = np.arange(self.bn//2, len(flux)-self.bn//2)
    else:
      raise(PE.PyAValError("Inappropriate length of flux array.\n" + \
                           "  Either len(template) or len(template)-bn+1 is accepted.", \
                           solution="Adjust the length of the input spectrum."))
    if self.w is None:
      raise(PE.PyAValError("There is no vector of singular values yet.", \
                           solution="Call `decompose` first."))
    w1 = 1.0/self.w
    indi = np.where(self.w < wlimit)[0]
    w1[indi] = 0.0
    b = np.dot(self.v, np.dot(np.diag(w1), np.dot(self.u.T, np.matrix(flux[validIndi]).T)))
    if not asarray:
      return b
    else:
      return b.getA()[::,0]
  
  def getRVAxis(self, binsize, refWvl):
    """
      Calculate a radial velocity axis for the broadening function.
      
      .. note:: Here, it is assumed that the broadening function
                refers to a spectrum in wavelength space.
      
      Parameters
      ----------
      binsize : float
          Size of spectral bins.
      refWvl : float
          The reference wavelength.
    
      Returns
      -------
      Radial velocity axis : array
          An array containing the radial velocity shift
          pertaining to the individual bins in the
          broadening function. The unit is km/s.
    """
    if self.bn is None:
      raise(PE.PyAOrderError("The length of the broadening function has not yet been specified.", \
                             solution="Call `decompose` first, where the length (bn) is specified."))
    result = -np.arange(self.bn) + float(self.bn//2)
    result *= binsize
    result /= refWvl
    result *= 299792.458
    return result
    
    
  
  def decompose(self, template, bn):
    """
      Carry out the SVD of the "design matrix".
      
      This method creates the "design matrix" by applying
      a bin-wise shift to the template and uses numpy's
      `svd` algorithm to carry out the decomposition.
      
      The design matrix, `des` is written in the form:
      "`des` = u * w * transpose(v)". The matrices des,
      w, and u are stored in homonymous attributes.
      
      Parameters
      ----------
      template : array or matrix
          The template with respect to which the broadening
          function shall be calculated.
      bn : int
          Width (number of elements) of the broadening function.
          Needs to be odd.
    """
    if bn % 2 != 1:
      raise(PE.PyAValError("`bn` needs to be an odd number."))
    self.bn = bn
    self.template = template.copy()
    self._createDesignMatrix(template, bn)
    # Get SVD deconvolution of the design matrix
    # Note that the `svd` method of numpy returns
    # v.H instead of v.
    self.u, self.w, v = np.linalg.svd(self.des, full_matrices=False)
    self.v = v.T
    